# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Instrument class"""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import partial
from typing import Callable, get_type_hints

from qililab.config import logger
from qililab.platform.components.bus_element import BusElement
from qililab.result import Result
from qililab.settings import Settings
from qililab.typings.enums import InstrumentName, Parameter
from qililab.typings.instruments.device import Device


class Instrument(BusElement, ABC):
    """Abstract base class declaring the necessary attributes
    and methods for the instruments.

    Args:
        settings (Settings): Class containing the settings of the instrument.
    """

    name: InstrumentName

    @dataclass
    class InstrumentSettings(Settings):
        """Contains the settings of an instrument.

        Args:
            alias (str): Alias of the instrument.
            firmware (str): Firmware version installed on the instrument.
        """

        alias: str
        firmware: str

    settings: InstrumentSettings  # a subtype of settings must be specified by the subclass
    device: Device

    class CheckParameterValueString:
        """Property used to check if the set parameter value is a string."""

        def __init__(self, method: Callable):
            self._method = method

        def __get__(self, obj, objtype):
            """Support instance methods."""
            return partial(self.__call__, obj)

        def __call__(self, ref: "Instrument", *args, **kwargs):
            """
            Args:
                method (Callable): Class method.

            Raises:
                ValueError: If value is neither a float or int.
            """
            if "value" not in kwargs:
                raise ValueError("'value' not specified to update instrument settings.")
            value = kwargs["value"]
            if not isinstance(value, str):
                raise ValueError(f"value must be a string. Current type: {type(value)}")
            return self._method(ref, *args, **kwargs)

    class CheckParameterValueBool:
        """Property used to check if the set parameter value is a bool."""

        def __init__(self, method: Callable):
            self._method = method

        def __get__(self, obj, objtype):
            """Support instance methods."""
            return partial(self.__call__, obj)

        def __call__(self, ref: "Instrument", *args, **kwargs):
            """
            Args:
                method (Callable): Class method.

            Raises:
                ValueError: If value is neither a float or int.
            """
            if "value" not in kwargs:
                raise ValueError("'value' not specified to update instrument settings.")
            value = kwargs["value"]
            if not isinstance(value, bool):
                raise ValueError(f"value must be a bool. Current type: {type(value)}")
            return self._method(ref, *args, **kwargs)

    class CheckParameterValueFloatOrInt:
        """Property used to check if the set parameter value is a float or int."""

        def __init__(self, method: Callable):
            self._method = method

        def __get__(self, obj, objtype):
            """Support instance methods."""
            return partial(self.__call__, obj)

        def __call__(self, ref: "Instrument", *args, **kwargs):
            """
            Args:
                method (Callable): Class method.

            Raises:
                ValueError: If value is neither a float or int.
            """
            if "value" not in kwargs:
                raise ValueError("'value' not specified to update instrument settings.")
            value = kwargs["value"]
            if not isinstance(value, float) and not isinstance(value, int):
                raise ValueError(f"value must be a float or an int. Current type: {type(value)}")
            if isinstance(value, int):
                # setting a float as type as expected
                kwargs["value"] = float(value)
            return self._method(ref, *args, **kwargs)

    class CheckDeviceInitialized:
        """Property used to check if the device has been initialized."""

        def __init__(self, method: Callable):
            self._method = method

        def __get__(self, obj, objtype):
            """Support instance methods."""
            return partial(self.__call__, obj)

        def __call__(self, ref: "Instrument", *args, **kwargs):
            """
            Args:
                method (Callable): Class method.

            Raises:
                AttributeError: If device has not been initialized.
            """
            if not hasattr(ref, "device") and (not args or not hasattr(args[0], "device")):
                raise AttributeError("Instrument Device has not been initialized")
            return self._method(ref, *args, **kwargs) if hasattr(ref, "device") else self._method(*args, **kwargs)

    def __init__(self, settings: dict):
        settings_class: type[self.InstrumentSettings] = get_type_hints(self).get("settings")  # type: ignore
        self.settings = settings_class(**settings)

    @CheckDeviceInitialized
    @abstractmethod
    def initial_setup(self):
        """Set initial instrument settings."""

    @CheckDeviceInitialized
    def setup(self, parameter: Parameter, value: float | str | bool, channel_id: int | None = None):
        """Set instrument settings parameter to the corresponding value

        Args:
            parameter (Parameter): settings parameter to be updated
            value (float | str | bool): new value
            channel_id (int | None): channel identifier of the parameter to update
        """
        raise ParameterNotFound(f"Could not find parameter {parameter} in instrument {self.name}")

    def get(self, parameter: Parameter, channel_id: int | None = None):  # pylint: disable=unused-argument
        """Get instrument parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None): Channel identifier of the parameter to update.
        """
        if hasattr(self.settings, parameter.value):
            return getattr(self.settings, parameter.value)
        raise ParameterNotFound(f"Could not find parameter {parameter} in instrument {self.name}")

    @CheckDeviceInitialized
    @abstractmethod
    def turn_on(self):
        """Turn on an instrument."""

    def acquire_result(self) -> Result | None:
        """Acquire results of the measurement.

        In some cases this method might do nothing."""

    def acquire_qprogram_results(self, acquisitions: list[str]) -> list[Result]:  # type: ignore[empty-body]
        """Acquire results of the measurement.

        In some cases this method might do nothing.

        Args:
            acquisitions (list[str]): A list of acquisitions names.

        Returns:
            list[Result]: The acquired results in chronological order.
        """

    @CheckDeviceInitialized
    @abstractmethod
    def reset(self):
        """Reset instrument settings."""

    @CheckDeviceInitialized
    @abstractmethod
    def turn_off(self):
        """Turn off an instrument."""

    @property
    def alias(self):
        """Instrument 'alias' property.

        Returns:
            str: settings.alias.
        """
        return self.settings.alias

    @property
    def firmware(self):
        """Instrument 'firmware' property.

        Returns:
            str: settings.firmware.
        """
        return self.settings.firmware

    def __str__(self):
        """String representation of an instrument."""
        return f"{self.alias}"

    def set_parameter(self, parameter: Parameter, value: float | str | bool, channel_id: int | None = None):
        """Sets the parameter of a specific instrument.

        Args:
            parameter (Parameter): parameter settings of the instrument to update
            value (float | str | bool): value to update
            channel_id (int | None, optional): instrument channel to update, if multiple. Defaults to None.

        Returns:
            bool: True if the parameter is set correctly, False otherwise
        """
        if not hasattr(self, "device"):
            raise ValueError(
                f"Instrument {self.name.value} is not connected and cannot set the new value: {value} to the parameter {parameter}."
            )
        if channel_id is None:
            logger.debug("Setting parameter: %s to value: %f", parameter.value, value)
        if channel_id is not None:
            logger.debug("Setting parameter: %s to value: %f in channel %d", parameter.value, value, channel_id)

        return self.setup(parameter=parameter, value=value, channel_id=channel_id)

    def get_parameter(self, parameter: Parameter, channel_id: int | None = None):
        """Gets the parameter of a specific instrument.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None, optional): Instrument channel, if multiple. Defaults to None.

        Returns:
            str | int | float | bool: Parameter value.
        """
        return self.get(parameter=parameter, channel_id=channel_id)


class ParameterNotFound(Exception):
    """Error raised when a parameter in an instrument is not found."""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"ParameterNotFound: {self.message}"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Class to interface with the local oscillator RohdeSchwarz SGS100A
"""
from dataclasses import dataclass

from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.instruments.signal_generator import SignalGenerator
from qililab.instruments.utils import InstrumentFactory
from qililab.typings import InstrumentName, RohdeSchwarzSGS100A
from qililab.typings.enums import Parameter


@InstrumentFactory.register
class SGS100A(SignalGenerator):
    """Rohde & Schwarz SGS100A class

    Args:
        name (InstrumentName): name of the instrument
        device (RohdeSchwarz_SGS100A): Instance of the qcodes SGS100A class.
        settings (SGS100ASettings): Settings of the instrument.
    """

    name = InstrumentName.ROHDE_SCHWARZ

    @dataclass
    class SGS100ASettings(SignalGenerator.SignalGeneratorSettings):
        """Contains the settings of a specific signal generator."""

    settings: SGS100ASettings
    device: RohdeSchwarzSGS100A

    @Instrument.CheckDeviceInitialized
    def setup(
        self,
        parameter: Parameter,
        value: float | str | bool,
        channel_id: int | None = None,
    ):
        """Set R&S dbm power and frequency. Value ranges are:
        - power: (-120, 25).
        - frequency (1e6, 20e9).
        """
        if parameter == Parameter.POWER:
            self.settings.power = float(value)
            self.device.power(self.power)
            return
        if parameter == Parameter.LO_FREQUENCY:
            self.settings.frequency = float(value)
            self.device.frequency(self.frequency)
            return
        if parameter == Parameter.RF_ON:
            value = bool(value)
            if value:
                self.turn_on()
            else:
                self.turn_off()
            return
        raise ParameterNotFound(f"Invalid Parameter: {parameter.value}")

    @Instrument.CheckDeviceInitialized
    def initial_setup(self):
        """performs an initial setup"""
        self.device.power(self.power)
        self.device.frequency(self.frequency)
        if self.rf_on:
            self.device.on()
        else:
            self.device.off()

    @Instrument.CheckDeviceInitialized
    def turn_on(self):
        """Start generating microwaves."""
        self.settings.rf_on = True
        self.device.on()

    @Instrument.CheckDeviceInitialized
    def turn_off(self):
        """Stop generating microwaves."""
        self.settings.rf_on = False
        self.device.off()

    @Instrument.CheckDeviceInitialized
    def reset(self):
        """Reset instrument."""



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Drag IQPair."""
from qililab.waveforms.drag_correction import DragCorrection
from qililab.waveforms.gaussian import Gaussian

from .iq_pair import IQPair


class Drag(IQPair):  # pylint: disable=too-few-public-methods
    """Drag pulse. This is a gaussian drive pulse with an IQ pair where the I channel corresponds to the gaussian wave
    and the Q is the drag correction, which corresponds to the derivative of the I channel times a ``drag_coefficient``.

    Args:
        amplitude (float): Maximum amplitude of the pulse.
        duration (int): Duration of the pulse (ns).
        num_sigmas (float): Sigma number of the gaussian pulse shape. Defines the width of the gaussian pulse.
        drag_coefficient (float): Drag coefficient that gives the DRAG its imaginary components.
    """

    def __init__(self, amplitude: float, duration: int, num_sigmas: float, drag_coefficient: float):
        waveform_i = Gaussian(amplitude=amplitude, duration=duration, num_sigmas=num_sigmas)
        waveform_q = DragCorrection(drag_coefficient=drag_coefficient, waveform=waveform_i)

        super().__init__(I=waveform_i, Q=waveform_q)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Load method used to load experiment and results data."""
import glob
import os
from pathlib import Path

import yaml

from qililab.constants import DATA, EXPERIMENT_FILENAME, RESULTS_FILENAME
from qililab.experiment.experiment import Experiment
from qililab.result.results import Results


def _get_last_created_path(folderpath: Path) -> Path:
    """Get the last created path for files in a directory"""
    files_list = glob.glob(os.path.join(folderpath, "*"))
    if not files_list:
        raise ValueError("No previous experiment data found.")
    path = max(files_list, key=os.path.getctime)
    return Path(path)


def _get_last_created_experiment_path() -> Path:
    """Get the last created path for the experiment"""
    folderpath = os.environ.get(DATA, None)
    if folderpath is None:
        raise ValueError("Environment variable DATA is not set.")
    last_daily_directory_path = _get_last_created_path(folderpath=Path(folderpath))
    return _get_last_created_path(folderpath=last_daily_directory_path)


def load(path: str | None = None, load_experiment: bool = False) -> tuple[Experiment | None, Results | None]:
    """Load Experiment and Results from yaml data.

    Args:
        path (str): Path to folder.

    Returns:
        Tuple(Experiment | None, Results | None): Return Experiment and Results objects, or None.
    """
    parsed_path = Path(path) if isinstance(path, str) else _get_last_created_experiment_path()
    experiment, results = None, None
    if load_experiment and os.path.exists(parsed_path / EXPERIMENT_FILENAME):
        with open(parsed_path / EXPERIMENT_FILENAME, mode="r", encoding="utf-8") as experiment_file:
            experiment = Experiment.from_dict(yaml.safe_load(stream=experiment_file))

    if os.path.exists(parsed_path / RESULTS_FILENAME):
        with open(parsed_path / RESULTS_FILENAME, mode="r", encoding="utf-8") as results_file:
            results = Results(**yaml.safe_load(stream=results_file))

    if experiment is not None and results is not None:
        experiment.results = results

    return experiment, results



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Circuit Transpiler class"""


import contextlib
from dataclasses import asdict

import numpy as np
from qibo import gates
from qibo.gates import Gate, M
from qibo.models import Circuit

from qililab.chip import Coupler, Qubit
from qililab.constants import RUNCARD
from qililab.instruments import AWG
from qililab.pulse import Pulse, PulseEvent, PulseSchedule
from qililab.settings.gate_event_settings import GateEventSettings
from qililab.typings.enums import Line
from qililab.utils import Factory

from .gate_decompositions import translate_gates
from .native_gates import Drag, Wait


class CircuitTranspiler:
    """Handles circuit transpilation. It has 3 accessible methods:
    - `circuit_to_native`: transpiles a qibo circuit to native gates (Drag, CZ, Wait, M) and optionally RZ if optimize=False (optimize=True by default)
    - `circuit_to_pulses`: transpiles a native gate circuit to a `PulseSchedule`
    - `transpile_circuit`: runs both of the methods above sequentially
    """

    def __init__(self, platform):  # type: ignore # ignore typing to avoid importing platform and causing circular imports
        self.platform = platform

    def transpile_circuit(self, circuits: list[Circuit]) -> list[PulseSchedule]:
        """Transpiles a list of qibo.models.Circuit to a list of pulse schedules.
        First translates the circuit to a native gate circuit and applies virtual Z gates and phase corrections for CZ gates.
        Then it converts the native gate circuit to a pulse schedule using calibrated settings from the runcard.

        Args:
            circuits (list[Circuit]): list of qibo circuits
        """
        native_circuits = (self.circuit_to_native(circuit) for circuit in circuits)
        return self.circuit_to_pulses(list(native_circuits))

    def circuit_to_native(self, circuit: Circuit, optimize: bool = True) -> Circuit:
        """Converts circuit with qibo gates to circuit with native gates

        Args:
            circuit (Circuit): circuit with qibo gates
            optimize (bool): optimize the transpiled circuit using otpimize_transpilation

        Returns:
            new_circuit (Circuit): circuit with transpiled gates
        """
        # init new circuit
        new_circuit = Circuit(circuit.nqubits)
        # add transpiled gates to new circuit, optimize if needed
        if optimize:
            gates_to_optimize = translate_gates(circuit.queue)
            new_circuit.add(self.optimize_transpilation(circuit.nqubits, ngates=gates_to_optimize))
        else:
            new_circuit.add(translate_gates(circuit.queue))

        return new_circuit

    def optimize_transpilation(self, nqubits: int, ngates: list[gates.Gate]) -> list[gates.Gate]:
        """Optimizes transpiled circuit by applying virtual Z gates.
        This is done by moving all RZ to the left of all operators as a single RZ. The corresponding cumulative rotation
        from each RZ is carried on as phase in all drag pulses left of the RZ operator.
        Virtual Z gates are also applied to correct phase errors from CZ gates.
        The final RZ operator left to be applied as the last operator in the circuit can afterwards be removed since the last
        operation is going to be a measurement, which is performed on the Z basis and is therefore invariant under rotations
        around the Z axis.
        This last step can also be seen from the fact that an RZ operator applied on a single qubit, with no operations carried
        on afterwards induces a phase rotation. Since phase is an imaginary unitary component, its absolute value will be 1
        independent on any (unitary) operations carried on it.

        Mind that moving an operator to the left is equivalent to applying this operator last so
        it is actually moved to the _right_ of Circuit.queue (last element of list).

        For more information on virtual Z gates, see https://arxiv.org/abs/1612.00858

        Args:
            nqubits (int) : number of qubits in the circuit
            ngates (list[gates.Gate]) : list of gates in the circuit

        Returns:
            list[gates.Gate] : list of re-ordered gates
        """
        supported_gates = ["rz", "drag", "cz", "wait", "measure"]
        new_gates = []
        shift = {qubit: 0 for qubit in range(nqubits)}
        for gate in ngates:
            if gate.name not in supported_gates:
                raise NotImplementedError(f"{gate.name} not part of native supported gates {supported_gates}")
            if isinstance(gate, gates.RZ):
                shift[gate.qubits[0]] += gate.parameters[0]
            # add CZ phase correction
            elif isinstance(gate, gates.CZ):
                gate_settings = self.platform.gates_settings.get_gate(name=gate.__class__.__name__, qubits=gate.qubits)
                control_qubit, target_qubit = gate.qubits
                corrections = next(
                    (
                        event.pulse.options
                        for event in gate_settings
                        if (
                            event.pulse.options is not None
                            and f"q{control_qubit}_phase_correction" in event.pulse.options
                        )
                    ),
                    None,
                )
                if corrections is not None:
                    shift[control_qubit] += corrections[f"q{control_qubit}_phase_correction"]
                    shift[target_qubit] += corrections[f"q{target_qubit}_phase_correction"]
                new_gates.append(gate)
            else:
                # if gate is drag pulse, shift parameters by accumulated Zs
                if isinstance(gate, Drag):
                    # create new drag pulse rather than modify parameters of the old one
                    gate = Drag(gate.qubits[0], gate.parameters[0], gate.parameters[1] - shift[gate.qubits[0]])

                # append gate to optimized list
                new_gates.append(gate)

        return new_gates

    def circuit_to_pulses(self, circuits: list[Circuit]) -> list[PulseSchedule]:  # pylint: disable=too-many-locals
        """Translates a list of circuits into a list of pulse sequences (each circuit to an independent pulse sequence)
        For each circuit gate we look up for its corresponding gates settings in the runcard (the name of the class of the circuit
        gate and the name of the gate in the runcard should match) and load its schedule of GateEvents.
        Each gate event corresponds to a concrete pulse applied at a certain time w.r.t the gate's start time and through a specific bus
        (see gates settings docstrings for more details).

        Measurement gates are handled in a slightly different manner. For a circuit gate M(0,1,2) the settings for each M(0), M(1), M(2)
        will be looked up and will be applied in sync. Note that thus a circuit gate for M(0,1,2) is different from the circuit sequence
        M(0)M(1)M(2) since the later will not be necessarily applied at the same time for all the qubits involved.

        Times for each qubit are kept track of with the dictionary `time`.
        The times at which each pulse is applied are padded if they are not multiples of the minimum clock time. This means that if min clock
        time is 4 and a pulse applied to qubit k lasts 17ns, the next pulse at qubit k will be at t=20ns

        Args:
            circuits (List[Circuit]): List of Qibo Circuit classes.

        Returns:
            list[PulseSequences]: List of :class:`PulseSequences` classes.
        """

        pulse_schedule_list: list[PulseSchedule] = []
        for circuit in circuits:
            pulse_schedule = PulseSchedule()
            time: dict[int, int] = {}  # init/restart time
            for gate in circuit.queue:
                # handle wait gates
                if isinstance(gate, Wait):
                    self._update_time(time=time, qubit=gate.qubits[0], gate_time=gate.parameters[0])
                    continue

                # Measurement gates need to be handled on their own because qibo allows to define
                # an M gate as eg. gates.M(*range(5))
                if isinstance(gate, M):
                    gate_schedule = []
                    gate_qubits = gate.qubits
                    for qubit in gate_qubits:
                        gate_schedule += self._gate_schedule_from_settings(M(qubit))

                # handle control gates
                else:
                    # extract gate schedule
                    gate_schedule = self._gate_schedule_from_settings(gate)
                    gate_qubits = self._get_gate_qubits(gate, gate_schedule)

                # process gate_schedule to pulses for both M and control gates
                # get total duration for the gate
                gate_time = self._get_total_schedule_duration(gate_schedule)
                # update time, start time is that of the qubit most ahead in time
                start_time = 0
                for qubit in gate_qubits:
                    start_time = max(self._update_time(time=time, qubit=qubit, gate_time=gate_time), start_time)
                # sync gate end time
                self._sync_qubit_times(gate_qubits, time=time)
                # apply gate schedule
                for gate_event in gate_schedule:
                    # find bus
                    bus = self.platform._get_bus_by_alias(gate_event.bus)  # pylint: disable=protected-access
                    # add control gate schedule
                    pulse_event = self._gate_element_to_pulse_event(
                        time=start_time, gate=gate, gate_event=gate_event, bus=bus
                    )
                    # pop first qubit from gate if it is measurement
                    # this is so that the target qubit for multiM gates is every qubit in the M gate
                    if isinstance(gate, M):
                        gate = M(*gate.qubits[1:])
                    # add event
                    pulse_schedule.add_event(pulse_event=pulse_event, port=bus.port, port_delay=bus.settings.delay)  # type: ignore

            for qubit in self.platform.chip.qubits:
                with contextlib.suppress(ValueError):
                    # If we find a flux port, create empty schedule for that port.
                    # This is needed because for Qblox instrument working in flux buses as DC sources, if we don't
                    # add an empty schedule its offsets won't be activated and the results will be misleading.
                    flux_port = self.platform.chip.get_port_from_qubit_idx(idx=qubit, line=Line.FLUX)
                    if flux_port is not None:
                        flux_bus = next((bus for bus in self.platform.buses if bus.port == flux_port), None)
                        if flux_bus and any(
                            isinstance(instrument, AWG) for instrument in flux_bus.system_control.instruments
                        ):
                            pulse_schedule.create_schedule(port=flux_port)

            pulse_schedule_list.append(pulse_schedule)

        return pulse_schedule_list

    def _gate_schedule_from_settings(self, gate: Gate) -> list[GateEventSettings]:
        """Gets the gate schedule. The gate schedule is the list of pulses to apply
        to a given bus for a given gate

        Args:
            gate (Gate): Qibo gate

        Returns:
            list[GateEventSettings]: schedule list with each of the pulses settings
        """

        gate_schedule = self.platform.gates_settings.get_gate(name=gate.__class__.__name__, qubits=gate.qubits)

        if not isinstance(gate, Drag):
            return gate_schedule

        # drag gates are currently the only parametric gates we are handling and they are handled here
        if len(gate_schedule) > 1:
            raise ValueError(
                f"Schedule for the drag gate is expected to have only 1 pulse but instead found {len(gate_schedule)} pulses"
            )
        drag_schedule = GateEventSettings(
            **asdict(gate_schedule[0])
        )  # make new object so that gate_schedule is not overwritten
        theta = self._normalize_angle(angle=gate.parameters[0])
        amplitude = drag_schedule.pulse.amplitude * theta / np.pi
        phase = self._normalize_angle(angle=gate.parameters[1])
        if amplitude < 0:
            amplitude = -amplitude
            phase = self._normalize_angle(angle=gate.parameters[1] + np.pi)
        drag_schedule.pulse.amplitude = amplitude
        drag_schedule.pulse.phase = phase
        return [drag_schedule]

    def _normalize_angle(self, angle: float):
        """Normalizes angle in range [-pi, pi].

        Args:
            angle (float): Normalized angle.
        """
        angle %= 2 * np.pi
        if angle > np.pi:
            angle -= 2 * np.pi
        return angle

    def _get_total_schedule_duration(self, schedule: list[GateEventSettings]) -> int:
        """Returns total time for a gate schedule. This is done by taking the max of (init + duration)
        for all the elements in the schedule

        Args:
            schedule (list[CircuitPulseSettings]): Schedule of pulses to apply

        Returns:
            int: Total gate time
        """
        time = 0
        for schedule_element in schedule:
            time = max(time, schedule_element.pulse.duration + schedule_element.wait_time)
        return time

    def _get_gate_qubits(self, gate: Gate, schedule: list[GateEventSettings] | None = None) -> list[int]:
        """Gets qubits involved in gate. This includes gate.qubits but also qubits which are targets of
        buses in the gate schedule

        Args:
            schedule (list[CircuitPulseSettings]): Gate schedule

        Returns:
            list[int]: list of qubits
        """

        schedule_qubits = (
            [
                target.qubit_index
                for schedule_element in schedule
                for target in self.platform._get_bus_by_alias(  # pylint: disable=protected-access
                    schedule_element.bus
                ).targets
                if isinstance(target, Qubit)
            ]
            if schedule is not None
            else []
        )

        gate_qubits = list(gate.qubits)

        return list(set(schedule_qubits + gate_qubits))  # converto to set and back to list to remove repeated items

    def _gate_element_to_pulse_event(
        self, time: int, gate: Gate, gate_event: GateEventSettings, bus  # type: ignore
    ) -> PulseEvent:
        """Translates a gate element into a pulse.

        Args:
            time (dict[int, int]): dictionary containing qubit indices as keys and current time (ns) as values
            gate (gate): circuit gate. This is used only to know the qubit target of measurement gates
            gate_event (GateEventSettings): gate event, a single element of a gate schedule containing information
            about the pulse to be applied
            bus (bus): bus through which the pulse is sent

        Returns:
            PulseEvent: pulse event corresponding to the input gate event
        """

        # copy to avoid modifying runcard settings
        pulse = gate_event.pulse
        pulse_shape_copy = pulse.shape.copy()
        pulse_shape = Factory.get(pulse_shape_copy.pop(RUNCARD.NAME))(**pulse_shape_copy)

        # handle measurement gates and target qubits for control gates which might have multi-qubit schedules
        if isinstance(gate, M):
            qubit = gate.qubits[0]
        # for couplers we don't need to set the target qubit
        elif isinstance(bus.targets[0], Coupler):
            qubit = None
        # handle control gates, target should be the qubit target of the bus
        else:
            qubit = next(target.qubit_index for target in bus.targets if isinstance(target, Qubit))

        return PulseEvent(
            pulse=Pulse(
                amplitude=pulse.amplitude,
                phase=pulse.phase,
                duration=pulse.duration,
                frequency=0,
                pulse_shape=pulse_shape,
            ),
            start_time=time + gate_event.wait_time + self.platform.gates_settings.delay_before_readout,
            pulse_distortions=bus.distortions,
            qubit=qubit,
        )

    def _update_time(self, time: dict[int, int], qubit: int, gate_time: int):
        """Creates new timeline if not already created and update time.

        Args:
            time (Dict[int, int]): Dictionary with the time of each qubit.
            qubit_idx (int): qubit index
            gate_time (int): total duration of the gate
        """
        if qubit not in time:
            time[qubit] = 0
        old_time = time[qubit]
        residue = (gate_time) % self.platform.gates_settings.minimum_clock_time
        if residue != 0:
            gate_time += self.platform.gates_settings.minimum_clock_time - residue
        time[qubit] += gate_time
        return old_time

    def _sync_qubit_times(self, qubits: list[int], time: dict[int, int]):
        """Syncs the time of the given qubit list

        Args:
            qubits (list[int]): qubits to sync
            time (dict[int,int]): time dictionary
        """
        max_time = max((time[qubit] for qubit in qubits if qubit in time), default=0)
        for qubit in qubits:
            time[qubit] = max_time



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import math
from collections import deque
from typing import Callable

import numpy as np
from qm import qua
from qm.qua import _dsl as qua_dsl
from qualang_tools.config.integration_weights_tools import convert_integration_weights

from qililab.qprogram.blocks import Average, Block, ForLoop, Loop, Parallel
from qililab.qprogram.blocks.infinite_loop import InfiniteLoop
from qililab.qprogram.operations import (
    Measure,
    Operation,
    Play,
    ResetPhase,
    SetFrequency,
    SetGain,
    SetPhase,
    Sync,
    Wait,
)
from qililab.qprogram.qprogram import QProgram
from qililab.qprogram.variable import Domain, Variable
from qililab.waveforms import IQPair, Square, Waveform

# mypy: disable-error-code="operator"


class _BusCompilationInfo:  # pylint: disable=too-few-public-methods
    def __init__(self):
        self.current_gain: float | qua.QuaVariableType | None = None


class _MeasurementCompilationInfo:  # pylint: disable=too-few-public-methods, too-many-instance-attributes
    def __init__(
        self,
        bus: str,
        variable_I: qua.QuaVariableType | None = None,
        variable_Q: qua.QuaVariableType | None = None,
        stream_I: qua_dsl._ResultSource | None = None,
        stream_Q: qua_dsl._ResultSource | None = None,
        stream_raw_adc: qua_dsl._ResultSource | None = None,
    ):
        self.bus: str = bus
        self.variable_I: qua.QuaVariableType | None = variable_I
        self.variable_Q: qua.QuaVariableType | None = variable_Q
        self.stream_I: qua_dsl._ResultSource | None = stream_I
        self.stream_Q: qua_dsl._ResultSource | None = stream_Q
        self.stream_raw_adc: qua_dsl._ResultSource | None = stream_raw_adc
        self.loops_iterations: list[int] = []
        self.average: bool = False


class MeasurementInfo:  # pylint: disable=too-few-public-methods
    """Class representing information about the measurements taking place."""

    def __init__(self, bus: str, result_handles: list[str]):
        self.bus: str = bus
        self.result_handles: list[str] = result_handles


class QuantumMachinesCompiler:  # pylint: disable=too-many-instance-attributes
    """A class for compiling QProgram to Quantum Machines hardware."""

    FREQUENCY_COEFF = 1e3
    PHASE_COEFF = 360.0
    VOLTAGE_COEFF = 2
    MINIMUM_TIME = 4

    def __init__(self):
        # Handlers to map each operation to a corresponding handler function
        self._handlers: dict[type, Callable] = {
            InfiniteLoop: self._handle_infinite_loop,
            Parallel: self._handle_parallel_loop,
            ForLoop: self._handle_for_loop,
            Loop: self._handle_loop,
            Average: self._handle_average,
            Measure: self._handle_measure,
            Play: self._handle_play,
            SetFrequency: self._handle_set_frequency,
            SetPhase: self._handle_set_phase,
            SetGain: self._handle_set_gain,
            ResetPhase: self._handle_reset_phase,
            Wait: self._handle_wait,
            Sync: self._handle_sync,
        }

        self._qprogram: QProgram
        self._bus_mapping: dict[str, str] | None
        self._qprogram_block_stack: deque[Block]
        self._qprogram_to_qua_variables: dict[Variable, qua.QuaVariableType]
        self._measurements: list[_MeasurementCompilationInfo]
        self._configuration: dict
        self._buses: dict[str, _BusCompilationInfo]

    def compile(
        self, qprogram: QProgram, bus_mapping: dict[str, str] | None = None
    ) -> tuple[qua.Program, dict, list[MeasurementInfo]]:
        """Compile QProgram to QUA's Program.

        Args:
            qprogram (QProgram): The QProgram to be compiled
            bus_mapping (dict[str, str] | None, optional): Optional mapping of bus names. Defaults to None.

        Returns:
            tuple[qua.Program, dict, list[MeasurementInfo]]: A tuple consisting of the generated QUA program, the generated configuration and a list of measurements' information.
        """

        def traverse(block: Block):
            self._qprogram_block_stack.append(block)
            for element in block.elements:
                handler = self._handlers.get(type(element))
                if not handler:
                    raise NotImplementedError(
                        f"{element.__class__} operation is currently not supported in Quantum Machines."
                    )
                if isinstance(element, (InfiniteLoop, ForLoop, Loop, Average, Parallel)):
                    with handler(element):
                        traverse(element)
                else:
                    handler(element)
            self._qprogram_block_stack.pop()

        self._qprogram = qprogram
        self._bus_mapping = bus_mapping

        self._qprogram_block_stack = deque()
        self._qprogram_to_qua_variables = {}
        self._measurements = []
        self._configuration = {
            "waveforms": {},
            "pulses": {},
            "integration_weights": {},
            "digital_waveforms": {"ON": {"samples": [(1, 0)]}, "OFF": {"samples": [(0, 0)]}},
            "elements": {},
        }

        self._populate_buses()

        with qua.program() as qua_program:
            # Declare variables
            self._declare_variables()
            # Recursive traversal to convert QProgram to QUA program.
            traverse(self._qprogram.body)
            # Stream Processing
            with qua.stream_processing():
                measurements = self._process_measurements()

        # Return a dictionary with bus names as keys and the compiled Sequence as values.
        return qua_program, self._configuration, measurements

    def _process_measurements(self):
        def _process_measurement(measurement: _MeasurementCompilationInfo, index: int):
            result_handles: list[str] = []

            def _process_stream(stream: qua_dsl._ResultSource, save_as: str) -> None:
                processing_stream: qua_dsl._ResultSource | qua_dsl._ResultStream = stream
                for loop_iteration in measurement.loops_iterations:
                    processing_stream = processing_stream.buffer(loop_iteration)
                if measurement.average:
                    processing_stream = processing_stream.average()
                processing_stream.save(save_as)
                result_handles.append(save_as)

            if measurement.stream_I is not None:
                _process_stream(measurement.stream_I, f"I_{index}")
            if measurement.stream_Q is not None:
                _process_stream(measurement.stream_Q, f"Q_{index}")
            if measurement.stream_raw_adc is not None:
                # TODO: Check if both inputs are available
                _process_stream(measurement.stream_raw_adc.input1(), f"adc1_{index}")
                _process_stream(measurement.stream_raw_adc.input2(), f"adc2_{index}")

            return result_handles

        measurements = [
            MeasurementInfo(bus=measurement.bus, result_handles=_process_measurement(measurement=measurement, index=i))
            for i, measurement in enumerate(self._measurements)
        ]

        return measurements

    def _declare_variables(self):
        for variable in self._qprogram.variables:
            if variable.domain in [Domain.Time, Domain.Frequency]:
                qua_variable = qua.declare(int)
            else:
                qua_variable = qua.declare(qua.fixed)
            self._qprogram_to_qua_variables[variable] = qua_variable

    def _populate_buses(self):
        """Map each bus in the QProgram to a BusCompilationInfo instance."""

        def collect_buses(block: Block):
            for element in block.elements:
                if isinstance(element, Block):
                    yield from collect_buses(element)
                if isinstance(element, Operation):
                    bus = getattr(element, "bus", None)
                    if bus:
                        yield self._bus_mapping[bus] if self._bus_mapping and bus in self._bus_mapping else bus

        buses = set(collect_buses(self._qprogram.body))
        self._configuration["elements"] = {bus: {"operations": {}} for bus in buses}
        self._buses = {bus: _BusCompilationInfo() for bus in buses}

    def _handle_infinite_loop(self, _: InfiniteLoop):
        return qua.infinite_loop_()

    def _handle_parallel_loop(self, element: Parallel):
        variables = []
        loops = []
        for loop in element.loops:
            qua_variable = self._qprogram_to_qua_variables[loop.variable]
            values = (
                np.arange(loop.start, loop.stop + loop.step / 2, loop.step)
                if isinstance(loop, ForLoop)
                else loop.values
            )
            if loop.variable.domain is Domain.Phase:
                values = values / self.PHASE_COEFF
            if loop.variable.domain is Domain.Frequency:
                values = (values * self.FREQUENCY_COEFF).astype(int)
            if loop.variable.domain is Domain.Time:
                values = np.maximum(values, self.MINIMUM_TIME).astype(int)

            variables.append(qua_variable)
            loops.append(values)
        return qua.for_each_(tuple(variables), tuple(loops))

    def _handle_for_loop(self, element: ForLoop):
        qua_variable = self._qprogram_to_qua_variables[element.variable]
        start, stop, step = element.start, element.stop, element.step
        if element.variable.domain is Domain.Phase:
            start, stop, step = start / self.PHASE_COEFF, stop / self.PHASE_COEFF, step / self.PHASE_COEFF
        if element.variable.domain is Domain.Frequency:
            start, stop, step = (
                int(start * self.FREQUENCY_COEFF),
                int(stop * self.FREQUENCY_COEFF),
                int(step * self.FREQUENCY_COEFF),
            )
        if element.variable.domain is Domain.Time:
            start = max(start, self.MINIMUM_TIME)

        to_positive = stop >= start
        if to_positive:
            return qua.for_(qua_variable, start, qua_variable <= stop, qua_variable + step)
        return qua.for_(qua_variable, start, qua_variable >= stop, qua_variable + step)

    def _handle_loop(self, element: Loop):
        qua_variable = self._qprogram_to_qua_variables[element.variable]
        values = element.values
        if element.variable.domain is Domain.Phase:
            values = values / self.PHASE_COEFF
        if element.variable.domain is Domain.Frequency:
            values = (values * self.FREQUENCY_COEFF).astype(int)
        if element.variable.domain is Domain.Time:
            values = np.maximum(values, self.MINIMUM_TIME).astype(int)
        return qua.for_each_(qua_variable, values)

    def _handle_average(self, element: Average):
        variable = qua.declare(int)
        return qua.for_(variable, 0, variable < element.shots, variable + 1)

    def _handle_set_frequency(self, element: SetFrequency):
        bus = self._bus_mapping[element.bus] if self._bus_mapping and element.bus in self._bus_mapping else element.bus
        frequency = (
            self._qprogram_to_qua_variables[element.frequency]
            if isinstance(element.frequency, Variable)
            else element.frequency * self.FREQUENCY_COEFF
        )
        qua.update_frequency(element=bus, new_frequency=frequency, units="mHz")

    def _handle_set_phase(self, element: SetPhase):
        bus = self._bus_mapping[element.bus] if self._bus_mapping and element.bus in self._bus_mapping else element.bus
        phase = (
            self._qprogram_to_qua_variables[element.phase]
            if isinstance(element.phase, Variable)
            else element.phase / self.PHASE_COEFF
        )
        qua.frame_rotation_2pi(phase, bus)

    def _handle_reset_phase(self, element: ResetPhase):
        bus = self._bus_mapping[element.bus] if self._bus_mapping and element.bus in self._bus_mapping else element.bus
        qua.reset_frame(bus)

    def _handle_set_gain(self, element: SetGain):
        bus = self._bus_mapping[element.bus] if self._bus_mapping and element.bus in self._bus_mapping else element.bus
        gain = self._qprogram_to_qua_variables[element.gain] if isinstance(element.gain, Variable) else element.gain
        # QUA doesn't have a method for setting the gain directly.
        # Instead, it uses an amplitude multiplication with `amp()`.
        # Thus, we store the current gain to multiply with in the next play instructions.
        self._buses[bus].current_gain = gain

    def _handle_play(self, element: Play):
        bus = self._bus_mapping[element.bus] if self._bus_mapping and element.bus in self._bus_mapping else element.bus
        waveform_I, waveform_Q = element.get_waveforms()
        waveform_variables = element.get_waveform_variables()
        duration = waveform_I.get_duration()

        gain = (
            qua.amp(self._buses[bus].current_gain * self.VOLTAGE_COEFF)
            if self._buses[bus].current_gain is not None
            else None
        )

        if not waveform_variables:
            waveform_I_name = self.__add_waveform_to_configuration(waveform_I)
            waveform_Q_name = self.__add_waveform_to_configuration(waveform_Q) if waveform_Q else None
            pulse_name = self.__add_or_update_control_pulse_to_configuration(waveform_I_name, waveform_Q_name, duration)
            operation_name = self.__add_pulse_to_element_operations(bus, pulse_name)
            pulse = operation_name * gain if gain is not None else operation_name
            qua.play(pulse, bus)

    def _handle_measure(
        self, element: Measure
    ):  # pylint: disable=too-many-locals, too-many-branches, too-many-statements
        bus = self._bus_mapping[element.bus] if self._bus_mapping and element.bus in self._bus_mapping else element.bus
        waveform_I, waveform_Q = element.get_waveforms()

        waveform_I_name = self.__add_waveform_to_configuration(waveform_I)
        waveform_Q_name = self.__add_waveform_to_configuration(waveform_Q)

        gain = (
            qua.amp(self._buses[bus].current_gain * self.VOLTAGE_COEFF)
            if self._buses[bus].current_gain is not None
            else None
        )

        variable_I, variable_Q, stream_I, stream_Q, stream_raw_adc = None, None, None, None, None

        if element.save_raw_adc:
            stream_raw_adc = qua.declare_stream(adc_trace=True)

        if element.weights is None:
            pulse_name = self.__add_or_update_measurement_pulse_to_configuration(
                waveform_I_name, waveform_Q_name, duration=waveform_I.get_duration(), integration_weights=[]
            )
            operation_name = self.__add_pulse_to_element_operations(bus, pulse_name)
            pulse = operation_name * gain if gain is not None else operation_name
            qua.measure(pulse, element.bus, stream_raw_adc)
        elif isinstance(element.weights, IQPair):
            variable_I = qua.declare(qua.fixed)
            stream_I = qua.declare_stream()
            weight_I = self.__add_integration_weight_to_configuration(element.weights.I, element.weights.Q)
            pulse_name = self.__add_or_update_measurement_pulse_to_configuration(
                waveform_I_name, waveform_Q_name, duration=waveform_I.get_duration(), integration_weights=[weight_I]
            )
            operation_name = self.__add_pulse_to_element_operations(bus, pulse_name)
            pulse = operation_name * gain if gain is not None else operation_name
            if element.demodulation:
                qua.measure(pulse, bus, stream_raw_adc, qua.demod.full(weight_I, variable_I, "out1"))
            else:
                qua.measure(pulse, bus, stream_raw_adc, qua.integration.full(weight_I, variable_I, "out1"))
            qua.save(variable_I, stream_I)
        elif isinstance(element.weights, tuple) and len(element.weights) == 2:
            variable_I = qua.declare(qua.fixed)
            variable_Q = qua.declare(qua.fixed)
            stream_I = qua.declare_stream()
            stream_Q = qua.declare_stream()
            weight_I = self.__add_integration_weight_to_configuration(element.weights[0].I, element.weights[0].Q)
            weight_Q = self.__add_integration_weight_to_configuration(element.weights[1].I, element.weights[1].Q)
            pulse_name = self.__add_or_update_measurement_pulse_to_configuration(
                waveform_I_name,
                waveform_Q_name,
                duration=waveform_I.get_duration(),
                integration_weights=[weight_I, weight_Q],
            )
            operation_name = self.__add_pulse_to_element_operations(bus, pulse_name)
            pulse = operation_name * gain if gain is not None else operation_name
            if element.demodulation:
                qua.measure(
                    pulse,
                    bus,
                    stream_raw_adc,
                    qua.demod.full(weight_I, variable_I, "out1"),
                    qua.demod.full(weight_Q, variable_Q, "out2"),
                )
            else:
                qua.measure(
                    pulse,
                    bus,
                    stream_raw_adc,
                    qua.integration.full(weight_I, variable_I, "out1"),
                    qua.integration.full(weight_Q, variable_Q, "out2"),
                )
            qua.save(variable_I, stream_I)
            qua.save(variable_Q, stream_Q)
        elif isinstance(element.weights, tuple) and len(element.weights) == 4:
            variable_I = qua.declare(qua.fixed)
            variable_Q = qua.declare(qua.fixed)
            stream_I = qua.declare_stream()
            stream_Q = qua.declare_stream()
            weight_A = self.__add_integration_weight_to_configuration(element.weights[0].I, element.weights[0].Q)
            weight_B = self.__add_integration_weight_to_configuration(element.weights[1].I, element.weights[1].Q)
            weight_C = self.__add_integration_weight_to_configuration(element.weights[2].I, element.weights[2].Q)  # type: ignore[misc]
            weight_D = self.__add_integration_weight_to_configuration(element.weights[3].I, element.weights[3].Q)  # type: ignore[misc]
            pulse_name = self.__add_or_update_measurement_pulse_to_configuration(
                waveform_I_name,
                waveform_Q_name,
                duration=waveform_I.get_duration(),
                integration_weights=[weight_A, weight_B, weight_C, weight_D],
            )
            operation_name = self.__add_pulse_to_element_operations(bus, pulse_name)
            pulse = operation_name * gain if gain is not None else operation_name
            if element.demodulation:
                qua.measure(
                    pulse,
                    bus,
                    stream_raw_adc,
                    qua.dual_demod.full(weight_A, "out1", weight_B, "out2", variable_I),
                    qua.dual_demod.full(weight_C, "out1", weight_D, "out2", variable_Q),
                )
            else:
                qua.measure(
                    pulse,
                    bus,
                    stream_raw_adc,
                    qua.dual_integration.full(weight_A, "out1", weight_B, "out2", variable_I),
                    qua.dual_integration.full(weight_C, "out1", weight_D, "out2", variable_Q),
                )
            qua.save(variable_I, stream_I)
            qua.save(variable_Q, stream_Q)

        measurement_info = _MeasurementCompilationInfo(bus, variable_I, variable_Q, stream_I, stream_Q, stream_raw_adc)
        for block in reversed(self._qprogram_block_stack):
            if isinstance(block, ForLoop):
                iterations = QuantumMachinesCompiler._calculate_iterations(block.start, block.stop, block.step)
                measurement_info.loops_iterations.append(iterations)
            if isinstance(block, Loop):
                iterations = len(block.values)
                measurement_info.loops_iterations.append(iterations)
            if isinstance(block, Parallel):
                iterations = min(
                    len(loop.values)
                    if isinstance(loop, Loop)
                    else QuantumMachinesCompiler._calculate_iterations(loop.start, loop.stop, loop.step)
                    for loop in block.loops
                )
                measurement_info.loops_iterations.append(iterations)
            if isinstance(block, Average):
                measurement_info.average = True

        self._measurements.append(measurement_info)

    def _handle_wait(self, element: Wait):
        bus = self._bus_mapping[element.bus] if self._bus_mapping and element.bus in self._bus_mapping else element.bus
        duration = (
            self._qprogram_to_qua_variables[element.duration]
            if isinstance(element.duration, Variable)
            else max(element.duration, self.MINIMUM_TIME)
        )
        qua.wait(duration, bus)

    def _handle_sync(self, element: Sync):
        if element.buses:
            buses = [
                self._bus_mapping[bus] if self._bus_mapping and bus in self._bus_mapping else bus
                for bus in element.buses
            ]
            qua.align(*buses)
        else:
            qua.align()

    def __add_pulse_to_element_operations(self, element: str, pulse: str):
        if pulse not in self._configuration["elements"][element]["operations"]:
            self._configuration["elements"][element]["operations"][pulse] = pulse
        return pulse

    def __add_or_update_control_pulse_to_configuration(
        self, waveform_I_name: str, waveform_Q_name: str | None, duration: int
    ):
        pulse_name = (
            f"control_{waveform_I_name}_{waveform_Q_name}_{duration}"
            if waveform_Q_name
            else f"control_{waveform_I_name}_{duration}"
        )
        if pulse_name not in self._configuration["pulses"]:
            configuration = {
                "operation": "control",
                "length": duration,
                "waveforms": {"I": waveform_I_name, "Q": waveform_Q_name}
                if waveform_Q_name
                else {"single": waveform_I_name},
            }
            self._configuration["pulses"][pulse_name] = configuration
        return pulse_name

    def __add_or_update_measurement_pulse_to_configuration(
        self, waveform_I_name: str, waveform_Q_name: str, duration: int, integration_weights: list[str]
    ):
        pulse_name = f"measurement_{waveform_I_name}_{waveform_Q_name}_{duration}"
        if pulse_name not in self._configuration["pulses"]:
            self._configuration["pulses"][pulse_name] = {
                "operation": "measurement",
                "length": duration,
                "waveforms": {
                    "I": waveform_I_name,
                    "Q": waveform_Q_name,
                },
                "integration_weights": {weight: weight for weight in integration_weights},
                "digital_marker": "ON",
            }
        else:
            self._configuration["pulses"][pulse_name]["integration_weights"].update(
                {weight: weight for weight in integration_weights}
            )
        return pulse_name

    def __add_integration_weight_to_configuration(self, cosine_part: Waveform, sine_part: Waveform):
        weight_name = f"{QuantumMachinesCompiler.__hash_waveform(cosine_part)}_{QuantumMachinesCompiler.__hash_waveform(sine_part)}"
        if weight_name not in self._configuration["integration_weights"]:
            self._configuration["integration_weights"][
                weight_name
            ] = QuantumMachinesCompiler.__integration_weight_to_config(cosine_part, sine_part)
        return weight_name

    def __add_waveform_to_configuration(self, waveform: Waveform):
        waveform_name = QuantumMachinesCompiler.__hash_waveform(waveform)
        if waveform_name not in self._configuration["waveforms"]:
            self._configuration["waveforms"][waveform_name] = QuantumMachinesCompiler.__waveform_to_config(waveform)
        return waveform_name

    @staticmethod
    def __hash_waveform(waveform: Waveform):
        attributes = [
            f"{key}: {(QuantumMachinesCompiler.__hash_waveform(value) if isinstance(value, Waveform) else str(value))}"
            for key, value in waveform.__dict__.items()
            if key != "duration"
        ]
        string_to_hash = f"{waveform.__class__.__name__}({','.join(attributes)})"
        hash_result = hashlib.md5(string_to_hash.encode("utf-8"), usedforsecurity=False)
        return hash_result.hexdigest()[:8]

    @staticmethod
    def __waveform_to_config(waveform: Waveform):
        if isinstance(waveform, Square):
            amplitude = waveform.amplitude / QuantumMachinesCompiler.VOLTAGE_COEFF
            return {"type": "constant", "sample": amplitude}

        envelope = waveform.envelope() / QuantumMachinesCompiler.VOLTAGE_COEFF
        return {"type": "arbitrary", "samples": envelope.tolist()}

    @staticmethod
    def __integration_weight_to_config(cosine_part: Waveform, sine_part: Waveform):
        cosine_envelope, sine_envelope = cosine_part.envelope(resolution=4), sine_part.envelope(resolution=4)
        converted_cosine_part = convert_integration_weights(integration_weights=cosine_envelope, N=len(cosine_envelope))
        converted_sine_part = convert_integration_weights(integration_weights=sine_envelope, N=len(sine_envelope))

        return {"cosine": converted_cosine_part, "sine": converted_sine_part}

    @staticmethod
    def _calculate_iterations(start: int | float, stop: int | float, step: int | float):
        if step == 0:
            raise ValueError("Step value cannot be zero")

        # Calculate the raw number of iterations
        raw_iterations = (stop - start + step) / step

        # If the raw number of iterations is very close to an integer, round it to that integer
        # This accounts for potential floating-point inaccuracies
        if abs(raw_iterations - round(raw_iterations)) < 1e-9:
            return round(raw_iterations)

        # Otherwise, if we're incrementing, take the ceiling, and if we're decrementing, take the floor
        return math.floor(raw_iterations) if step > 0 else math.ceil(raw_iterations)

    @staticmethod
    def merge_configurations(origin: dict, new: dict):
        """Recursively merge `new` into `origin`."""
        for key, value in new.items():
            if key in origin and isinstance(origin[key], dict) and isinstance(value, dict):
                QuantumMachinesCompiler.merge_configurations(origin[key], value)
            else:
                origin[key] = value
        return origin



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains a pre-defined version of a flipping sequence experiment."""
import numpy as np
from qibo.gates import RX, M, X
from qibo.models import Circuit

from qililab.platform import Platform
from qililab.typings import ExperimentOptions, ExperimentSettings, Parameter
from qililab.utils import Loop

from .experiment_analysis import ExperimentAnalysis
from .fitting_models import Cos


class FlippingSequence(ExperimentAnalysis, Cos):
    """Class used to run a flipping sequence experiment on the given qubit.

    This experiment creates multiple circuits, each of which uses an RX(pi/2) gate to send the qubit to the equator of
    the Bloch sphere, and then applies N R(2pi) gates to flip the qubit around the x axis N times. Given that the
    amplitude of the X gate is not well calibrated, the qubit will move towards the ground or the excited state.

    Args:
        platform (Platform): platform used to run the experiment
        qubit (int): qubit index used in the experiment
        loop_values (numpy.ndarray): array of values to loop through in the experiment, which determines the number of flips
        repetition_duration (int, optional): duration of a single repetition in nanoseconds. Defaults to 10000.
        hardware_average (int, optional): number of repetitions used to average the result. Defaults to 10000.
    """

    def __init__(
        self,
        platform: Platform,
        qubit: int,
        loop_values: np.ndarray,
        repetition_duration=10000,
        hardware_average=10000,
    ):
        # Define circuits used in this experiment
        loop = Loop(alias="N", parameter=Parameter.NUM_FLIPS, values=loop_values)
        circuits = []
        for n in loop.values:
            circuit = Circuit(1)
            circuit.add(RX(qubit, theta=np.pi / 2))
            for _ in range(n):
                circuit.add(X(qubit))
                circuit.add(X(qubit))
            circuit.add(M(qubit))
            circuits.append(circuit)

        _, control_bus, readout_bus = platform._get_bus_by_qubit_index(qubit)

        experiment_options = ExperimentOptions(
            name="Flipping Sequence",
            settings=ExperimentSettings(repetition_duration=repetition_duration, hardware_average=hardware_average),
        )

        # Initialize experiment
        super().__init__(
            platform=platform,
            circuits=circuits,
            options=experiment_options,
            control_bus=control_bus,
            readout_bus=readout_bus,
            experiment_loop=loop,
        )



"""Tests for the load function."""
from copy import deepcopy
from pathlib import Path
from unittest.mock import MagicMock, patch

from qililab import load
from qililab.utils import Loop
from tests.data import experiment, results_one_loops, results_two_loops


@patch("qililab.utils.load_data.os.path.exists", side_effect=lambda path: path == Path("results.yml"))
@patch("qililab.utils.load_data.open")
@patch("qililab.utils.load_data.yaml.safe_load", return_value=deepcopy(results_two_loops))
def test_load_results_two_loops(mock_load: MagicMock, mock_open: MagicMock, mock_os: MagicMock):
    """Test load Results class."""
    _, result = load(path="", load_experiment=True)
    mock_load.assert_called_once()
    mock_open.assert_called_once()
    mock_os.assert_called()
    assert result is not None
    assert result.loops is not None
    assert isinstance(result.loops[0], Loop)


@patch("qililab.utils.load_data.os.path.exists", side_effect=lambda path: path == Path("results.yml"))
@patch("qililab.utils.load_data.open")
@patch("qililab.utils.load_data.yaml.safe_load", return_value=deepcopy(results_one_loops))
def test_load_results_one_loop(mock_load: MagicMock, mock_open: MagicMock, mock_os: MagicMock):
    """Test load Results class."""
    _, result = load(path="", load_experiment=True)
    mock_load.assert_called_once()
    mock_open.assert_called_once()
    mock_os.assert_called()
    assert result is not None
    assert result.loops is not None
    assert isinstance(result.loops[0], Loop)


@patch("qililab.utils.load_data.os.path.exists", side_effect=lambda path: path == Path("experiment.yml"))
@patch("qililab.utils.load_data.open")
@patch("qililab.utils.load_data.yaml.safe_load", return_value=deepcopy(experiment))
def test_load_experiment(mock_load: MagicMock, mock_open: MagicMock, mock_os: MagicMock):
    """Test load Experiment class."""
    exp, _ = load(path="", load_experiment=True)
    mock_load.assert_called_once()
    mock_open.assert_called_once()
    mock_os.assert_called()
    assert exp is not None
    assert exp.options is not None
    assert exp.options.loops is not None
    assert isinstance(exp.options.loops[0], Loop)


@patch("qililab.utils.load_data.os.path.exists", side_effect=lambda _: True)
@patch("qililab.utils.load_data.open", side_effect=lambda filepath, *args, **kwargs: filepath)
@patch(
    "qililab.utils.load_data.yaml.safe_load",
    side_effect=lambda stream: deepcopy(results_one_loops if stream.name.endswith("results.yml") else experiment),
)
def test_load_experiment_and_results(mock_load: MagicMock, mock_open: MagicMock, mock_os: MagicMock):
    """Test load Experiment class."""
    exp, results = load(path="", load_experiment=True)
    mock_load.assert_called()
    mock_open.assert_called()
    mock_os.assert_called()
    assert exp is not None
    assert exp.options is not None
    assert exp.options.loops is not None
    assert isinstance(exp.options.loops[0], Loop)
    assert exp.results is results
    assert results is not None
    assert results.loops is not None
    assert isinstance(results.loops[0], Loop)


@patch("qililab.utils.load_data.glob.glob", return_value=["one", "two"])
@patch("qililab.utils.load_data.os.path.getctime", return_value=0)
def test_load_without_path(mock_cttime: MagicMock, mock_glob: MagicMock):
    """Test load without path."""
    exp, results = load(load_experiment=True)
    mock_cttime.assert_called()
    mock_glob.assert_called()
    assert exp is None
    assert results is None



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=anomalous-backslash-in-string
"""Automatic-calibration Controller module, which works with notebooks as nodes."""
from datetime import datetime, timedelta

import networkx as nx

from qililab.calibration.calibration_node import CalibrationNode
from qililab.config import logger
from qililab.data_management import build_platform, save_platform
from qililab.platform.platform import Platform


class CalibrationController:
    """Controls the automatic calibration sequence.

    **Usage:**
        - To calibrate the full graph, use the ``run_automatic_calibration()`` method.
        - To target and ensure a node works, use the ``maintain(node)`` method.

    |

    **Graph Structure:**

    In the graph, directions should be given by `nodes` pointing to their next `dependents` (natural time flow for calibration).
    This defines our `starts` and `ends` of the calibration:

    - `starts`: `Roots` (``in_degree=0``, no arrows pointing in, no `dependencies`) of the graph, where all arrows leave the node.

    - `ends`: `Leaves` (``out_degree=0``, no arrows pointing out, no `dependants`) of the graph, where all arrows enter the node.

    .. code-block:: python

        #                     /--> 2 -->\\
        #                    /     |     \\
        #    (start, root)  0      |      3 --> 4 (end, leave)
        #                    \     v     /
        #                     \--> 1 -->/

    .. note::

        Find information about the automatic-calibration workflow, in the examples below.

    Args:
        calibration_graph (nx.DiGraph): The calibration (directed acyclic) graph, where each node is a ``string`` corresponding to a ``CalibrationNode.node_id``. Directions should be given
            by `nodes` pointing to their next `dependents` (natural time flow for calibration), defining our `starts` and `ends` of the calibration as the `roots` (``in_degree=0``) and `leaves`
            (``out_degree=0``) of the graph.

        node_sequence (dict[str, CalibrationNode]): Mapping for the nodes of the graph, from strings into the actual initialized nodes.
        runcard (str): The runcard path, containing the serialized platform where the experiments will be run.

    Examples:

        **Calibration Workflow:**

        The calibration process is structured into three levels of methods:

        1. **Highest Level Method**: The ``run_automatic_calibration()`` method finds all the end nodes of the graph (`leaves`, those without further `dependents`) and runs ``maintain()`` on them.

        2. **Mid-Level Methods**: ``maintain()`` and ``diagnose()``.
            - ``maintain(node)`` starts from the `roots` that ``node`` depends on, and moves forwards (`dependency -> dependant`) until ``node``, checking the last time executions and data at each step. If a problem (``bad_data``) is found, it calls ``diagnose()`` to solve it.
            - ``diagnose(node)`` does more strict checks, fixing inaccuracies in the system's state to allow ``maintain()`` to continue. It works in reverse, starting from the problematic (``bad_data``) node, it goes back (`dependency <- dependant`) until it finds the origin of the problem.

        3. **Low-Level Methods**: ``check_state()``, ``check_data()``, and ``calibrate()`` are the methods you would be calling during this process to interact with the ``nodes``.

        |

        Finally, ``run_automatic_calibration()`` is designed to start acquiring data or calibrating in the optimal location of the graph to avoid extra work:

        - If node A has been calibrated very recently (before the ``drift_timeout`` of the :class:`.CalibrationNode`), it would be a waste of resources to check its data, so ``check_state()`` makes ``maintain()`` skip it.
        - If node A depends on node B, before calibrating node A, we check the data of node B. Calibrating A would be a waste of resources if we were doing so based on faulty data, so it goes to its dependencies first.

        .. note:: Find more information about the automatic calibration workflow at https://arxiv.org/abs/1803.03226.

        ----------

        **Dangerous Behaviors:**

        Note that depending on your ``CalibrationController`` construction, you can have dangerous behaviors in the workflow. You need to watch out for:

        - If you give bad ``comparison_thresholds`` or have bad ``comparison_models``, returning ``out_of_spec`` when you actually have ``bad_data`` or the other way around, will make ``diagnose()`` not being able to work properly, since for it to work, you need to have a single layer node separation (``out_of_spec``) between the ``in_spec`` and the ``bad_data`` nodes.
        - If you give too long ``drift_timeout``'s, since ``maintain()`` will assume the node is 100% working, you will go to further dependents when maybe you shouldn't without recalibrating. In the end ``diagnose()`` saves the day here, since it doesn't do ``check_state()``, although it will be less efficient.

        ----------

        **Practical example:**

        To create two linked nodes twice, for two different qubits, and pass them to a :class:`.CalibrationController` and run a ``maintain()``, you need:

        .. code-block:: python

            import numpy as np
            import networkx as nx

            from qililab.calibration import CalibrationController, CalibrationNode, norm_root_mean_sqrt_error

            # GRAPH CREATION AND NODE MAPPING (key = name in graph, value = node object):
            nodes = {}
            G = nx.DiGraph()
            first, second = [], []

            # CREATE NODES:
            for qubit in [0, 1]:
                first[qubit] = CalibrationNode(
                    nb_path="notebooks/first.ipynb",
                    qubit_index=qubit,
                    in_spec_threshold=4,
                    bad_data_threshold=8,
                    comparison_model=norm_root_mean_sqrt_error,
                    drift_timeout=1800.0,
                )
                nodes[first[qubit].node_id] = first[qubit]

                second[qubit] = CalibrationNode(
                    nb_path="notebooks/second.ipynb",
                    qubit_index=qubit,
                    in_spec_threshold=2,
                    bad_data_threshold=4,
                    comparison_model=norm_root_mean_sqrt_error,
                    drift_timeout=1.0,
                    sweep_interval=np.arange(start=0, stop=19, step=1),
                )
                nodes[second[qubit].node_id] = second[qubit]

                # GRAPH BUILDING (1 --> 2):
                G.add_edge(first[qubit].node_id, second[qubit].node_id)

            # CREATE CALIBRATION CONTROLLER:
            controller = CalibrationController(node_sequence=nodes, calibration_graph=G, runcard=path_runcard)

            ### WORKFLOW TO DO:
            controller.maintain(second[1]) # maintain second node for qubit 1

        .. note::

            Find information about how these nodes and their notebooks need to be in the :class:`CalibrationNode` class documentation.

            There you will also find the above code, but without defining ``first`` and ``second`` as lists.
    """

    def __init__(self, calibration_graph: nx.DiGraph, node_sequence: dict[str, CalibrationNode], runcard: str):
        if not nx.is_directed_acyclic_graph(calibration_graph):
            raise ValueError("The calibration graph must be a Directed Acyclic Graph (DAG).")

        self.calibration_graph: nx.DiGraph = calibration_graph
        """The calibration (directed acyclic) graph. Where each node is a ``string`` corresponding to a ``CalibrationNode.node_id``.

        Directions should be given by `nodes` pointing to their next `dependents` (natural time flow for calibration),
        defining our `starts` and `ends` of the calibration, as:

        - `starts`: `roots` (``in_degree=0``, no arrows pointing in, no `dependencies`) of the graph, where all arrows leave the node.

        - `ends`: `leaves` (``out_degree=0``, no arrows pointing out, no `dependants`) of the graph, where all arrows enter the node.

        .. code-block:: python

            #                     /--> 2 -->\\
            #                    /     |     \\
            #    (start, root)  0      |      3 --> 4 (end, leave)
            #                    \     v     /
            #                     \--> 1 -->/
        """

        self.node_sequence: dict[str, CalibrationNode] = node_sequence
        """Mapping for the nodes of the graph, from strings into the actual initialized nodes (dict)."""

        self.runcard: str = runcard
        """The runcard path, containing the serialized platform where the experiments will be run (str)."""

        self.platform: Platform = build_platform(runcard)
        """The initialized platform, where the experiments will be run (Platform)."""

    def run_automatic_calibration(self) -> dict[str, dict]:
        """Runs the full automatic calibration procedure and retrieves the final set parameters and achieved fidelities dictionaries.

        This is the primary interface for our calibration procedure and the highest level algorithm, which finds all the end nodes of the graph
        (`leaves`, those without further `dependents`) and runs ``maintain()`` on them.

        Returns:
            dict[str, dict]: Dictionary for the last set parameters and the last achieved fidelities. It contains two dictionaries (dict[tuple, tuple]) in the keys:
                - "set_parameters": Set parameters dictionary, with the key and values being tuples containing, in this order:
                    - key: (``str``: parameter name, ``str``: bus alias, int: qubit).
                    - value: (``float``: parameter value, ``str``: ``node_id`` where computed, ``datetime``: updated time).

                - "fidelities": Fidelities dictionary, with the key and values being tuples containing, in this order:
                    - key: (``str``: parameter name, ``int``: qubit).
                    - value: (``float``: parameter value, ``str``: node_id where computed, ``datetime``: updated time).
        """
        highest_level_nodes = [node for node, out_degree in self.calibration_graph.out_degree() if out_degree == 0]

        for n in highest_level_nodes:
            self.maintain(self.node_sequence[n])

        logger.info(
            "#############################################\n"
            "Automatic calibration completed successfully!\n"
            "#############################################\n"
        )

        return {"set_parameters": self.get_last_set_parameters(), "fidelities": self.get_last_fidelities()}

    def maintain(self, node: CalibrationNode) -> None:
        """Calls all the necessary subroutines (``check_state()``, ``check_data()`` and ``calibrate()``) in the respective dependencies to get a node
        in spec. Maintain contains the main workflow for our calibration procedure, and it's what is called from ``run_automatic_calibration()`` into
        each of the end ``nodes`` (leaves) of our graph.

        The algorithm starts from the `roots` that the passed ``node`` depends on, and moves forwards (`dependency -> dependant`) up to ``node``,
        at each step, first efficiently checking the last time executions (``check_state()``) until some fails, and then passes to check the data
        (``check_data()``, less efficient). If a problematic node is found, meaning, it doesn't pass ``check_state()`` and gives
        ``bad_data`` in ``check_data()``, then ``maintain()`` calls ``diagnose()`` for solving it.

        During all this process, the mid-level methods would be calling these low-level methods, that act on the nodes:
        ``calibrate()``, ``check_status()`` and ``check_data()``.

        Finally, ``maintain`` is designed to start acquiring data or calibrating in the optimal location of the graph, to avoid extra work:

        - if node A has been calibrated very recently (before the ``drift_timeout`` of the :class:`.CalibrationNode`), it would be waste of resources to check its data, so ``check_state()`` makes ``maintain()`` skip it.
        - if node A depends on node B, before calibrating node A we check the data of node B, calibrating A would be a waste of resources if we were doing so based on faulty data, so it goes to its dependencies first.

        |

        Note that due to this, we can have dangerous behaviours in ``maintain()``, for example:

        - if you give bad ``comparison_thresholds`` or have bad ``comparison_models``, returning ``out_of_spec`` when you actually have ``bad_data`` or the other way around, making ``diagnose()`` not being able to work properly. Since for ``diagnose()`` to work, you need to have a single layer node separation (``out_of_spec``) between the ``in_spec`` and the ``bad_data`` nodes.
        - if you give too long ``drift_timeout``'s, since ``maintain()`` will assume the node is 100% working, you will go to further dependants, when maybe you shouldn't without recalibrating (in the end ``diagnose()`` saves the day here, since it doesn't do ``check_state()``, but in a less efficient way).

        .. note:: Find more information about the ``maintain()`` idea at https://arxiv.org/abs/1803.03226.

        Args:
            node (CalibrationNode): The node where we want to start the algorithm on, getting it in spec. Normally you would want
                this node to be the furthest node in the calibration graph.
        """
        logger.info("Maintaining %s.\n", node.node_id)
        # Recursion over all the nodes that the current node depends on.
        for n in self._dependencies(node):
            logger.info("Maintaining %s from maintain(%s).\n", n.node_id, node.node_id)
            self.maintain(n)

        if self.check_state(node):
            return

        result = self.check_data(node)
        if result == "in_spec":
            return
        if result == "bad_data":
            for n in self._dependencies(node):
                logger.info("Diagnosing %s from maintain(%s).\n", n.node_id, node.node_id)
                self.diagnose(n)

        self.calibrate(node)
        self._update_parameters(node)

    def diagnose(self, node: CalibrationNode) -> bool:
        """Checks the data of all the dependencies of a node, until it finds the root of the problem with their data.

        This is a method called by ``maintain()`` in the special case that its call of ``check_data()`` finds bad data.

        ``Diagnose()`` workflow works in reverse, starting from the problematic (``bad_data``) node, it goes back (`dependency <- dependant`)
        until it finds the origin of the problem (the first and only ``out_of_spec`` of that path, since the previous will be ``in_spec`` and the
        followings in ``bad_data``).

        ``Maintain()`` assumes that our knowledge of the state of the system matches the actual state of the system: if we knew a node would
        return bad data, we wouldn't bother ``calibrating`` it. The fact that ``check_data()`` returns ``bad_data`` means that that's not the
        case, our knowledge of the systems's state is inaccurate. That why ``diagnose()`` does more strict checks, fixing inaccuracies in our
        knowledge of the system's state, to allow ``maintain()`` to continue.

        Finally mention, two important thing to have in mind:

        - if you give bad ``comparison_thresholds`` or have bad ``comparison_models``, which return ``out_of_spec`` when you actually have ``bad_data`` or the other way around, it will make your full calibration fail. Since for ``diagnose()`` to work, you need to have a single node separation (``out_of_spec``) between the ``in_spec`` and the ``bad_data`` ones.
        - if you have wrong ``drift_timeout``, the algorithm will be slower, but in the end ``diagnose()`` saves the day, achieving the full calibration, since it doesn't do ``check_state()``.

        .. note:: Find more information about the ``diagnose()`` idea at https://arxiv.org/abs/1803.03226.

        Args:
            node (CalibrationNode): The node where we want to start the algorithm.

        Returns:
            bool: True is there have been recalibrations, False otherwise. The return value is only used by recursive calls.
        """
        logger.info("diagnosing %s.\n", node.node_id)
        result = self.check_data(node)

        # in spec case
        if result == "in_spec":
            return False

        # bad data case
        recalibrated = []
        if result == "bad_data":
            recalibrated = [self.diagnose(n) for n in self._dependencies(node)]
            logger.info("Dependencies diagnoses of %s: %s\n", node.node_id, str(recalibrated))
        # If not empty and only filled with False's (not any True).
        if recalibrated != [] and not any(recalibrated):
            return False

        # calibrate
        self.calibrate(node)
        self._update_parameters(node)
        return True

    def check_state(self, node: CalibrationNode) -> bool:
        """Checks if the node's drift timeouts have passed since the last calibration or data check.

        These timeouts represent how long it usually takes for the parameters to drift, specified by the user.

        Conditions for ``check_state()`` to pass:
            - The cal has had ``check_data()`` or ``calibrate()`` pass within the timeout period.
            - The cal has not failed ``calibrate()`` without resolution.
            - No dependencies have been recalibrated since the last time ``check_data()`` or ``calibrate()`` was run on this cal.
            - All dependencies pass ``check_state()``.

        .. note:: Find more information about the ``check_state()`` idea at https://arxiv.org/abs/1803.03226.

        Args:
            node (CalibrationNode): The node whose state needs to be checked.

        Returns:
            bool: True if the parameter's drift timeout has not yet expired, False otherwise.
        """
        logger.info('Checking state of node "%s".\n', node.node_id)

        # Get the list of the dependencies that have been calibrated before this node, all of them should be True
        dependencies_timestamps_previous = [
            n.previous_timestamp < node.previous_timestamp for n in self._dependencies(node)
        ]
        # Check if something hapened and the timestamp could not be setted properly and the rest of conditions
        if node.previous_timestamp is None or not all(
            dependencies_timestamps_previous
        ):  # or not all(dependencies_status)
            logger.info("check_state of %s: False.\n", node.node_id)
            return False
        logger.info(
            "check_state of %s: %r.\n",
            node.node_id,
            (not self._is_timeout_expired(node.previous_timestamp, node.drift_timeout)),
        )
        return not self._is_timeout_expired(node.previous_timestamp, node.drift_timeout)

    def check_data(self, node: CalibrationNode) -> str:
        """Checks if the parameters found in the last calibration are still valid, doing a reduced execution of the notebook.

        To do this, ``check_data()`` runs the experiment only in a few points, randomly chosen within the ``sweep_interval``,
        and compares the results with the data obtained in the same points when the experiment was last ``calibrate()`` with the
        entire ``sweep_interval``.

        The comparison is done with the model is indicated by the ``comparison_model`` attribute of :class:`.CalibrationNode`,
        which returns a value indicating how well the data fits the model.

        This comparison is then classified as ``in_spec`` (still valid), ``out_of_spec`` (drifted, but close) or ``bad_data``
        (noise/doesn't follow the desired fit, or no previous execution) given the ``in_spec_threshold`` and ``bad_data_threshold``
        attributes of :class:`.CalibrationNode`.

        .. note:: Find more information about the ``check_data()`` idea at https://arxiv.org/abs/1803.03226.

        Args:
            node (CalibrationNode): The node whose parameters need to be checked.

        Returns:
            str: The status of the data, depending on how the current results compare with the obtained during the last full calibration.

            Concretely, depending on the provided thresholds and comparison models, it will return:

            - ``in_spec`` if the results are still acceptable to use.
            - ``out_of_spec`` if the results have drifted are not acceptable enough, but they are close, and still follow the desired fit.
            - ``bad_data`` if the results don't follow the desired fit, or are noisy, which should happen when dependencies have drifted. Or also if there are no previous executions.
        """
        # pylint: disable=protected-access

        logger.info('Checking data of node "%s".\n', node.node_id)
        timestamp = node.run_node(check=True)

        # Comparison and obtained parameters:
        comparison_outputs = node.previous_output_parameters
        obtained_outputs = node.output_parameters

        # If no previous result, return bad_data:
        comparison_result = "bad_data"

        # Do comparison and return the result:
        # (obtained_outputs should never be None, since we just run the notebook)
        if comparison_outputs is not None and obtained_outputs is not None:
            # Get comparison from last notebook and the new obtained parameters.
            compar_params = comparison_outputs["check_parameters"]
            obtain_params = obtained_outputs["check_parameters"]
            logger.info("obtained: %s ", str(obtain_params))
            logger.info("comparison: %s", str(compar_params))

            comparison_number = self._obtain_comparison(node, obtain_params, compar_params)

            if comparison_number <= node.in_spec_threshold:
                comparison_result = "in_spec"

            elif comparison_number <= node.bad_data_threshold:
                comparison_result = "out_of_spec"

        # Do the necessary following changes:
        logger.info("check_data of %s: %s.\n", node.node_id, comparison_result)
        node._add_string_to_checked_nb_name(comparison_result, timestamp)
        node.previous_output_parameters = node.output_parameters
        return comparison_result

    def calibrate(self, node: CalibrationNode) -> None:
        """Runs a node's experiment on its default values of the ``sweep_interval``.

        This method is responsible for calibrating a node to bring it within spec. The calibration process is node-specific
        and depends on the notebook implementation and the calibration parameters.

        .. note:: Find more information about the ``calibrate()`` idea at https://arxiv.org/abs/1803.03226.

        Args:
            node (CalibrationNode): The node where the calibration experiment is run.
        """
        logger.info('Calibrating node "%s".\n', node.node_id)
        node.previous_timestamp = node.run_node()
        node._add_string_to_checked_nb_name("calibrated", node.previous_timestamp)  # pylint: disable=protected-access

    def _update_parameters(self, node: CalibrationNode) -> None:
        """Updates the node parameters value in the platform, after a calibration.

        If the node does not have an associated parameter, or the parameter attribute of the node is None,
        this function does nothing.

        Args:
            node (CalibrationNode): The node which parameters need to be updated in the platform.
        """
        if node.output_parameters is not None and "platform_params" in node.output_parameters:
            for bus_alias, qubit, param_name, param_value in node.output_parameters["platform_params"]:
                logger.info(
                    "Platform updated with: (bus: %s, q: %s, %s, %f).", bus_alias, qubit, param_name, param_value
                )
                self.platform.set_parameter(alias=bus_alias, parameter=param_name, value=param_value, channel_id=qubit)

            save_platform(self.runcard, self.platform)

    def get_last_set_parameters(self) -> dict[tuple, tuple]:
        """Retrieves the last set parameters of the graph.

        Returns:
            dict[tuple, tuple]: Set parameters dictionary, with the key and values being tuples containing, in this order:
                - ``key``: (``str``: parameter name, ``str``: bus alias, ``int``: qubit).
                - ``value``: (``float``: parameter value, ``str``: node_id where computed, ``datetime``: updated time).
        """
        parameters: dict[tuple, tuple] = {}
        for node in self.node_sequence.values():
            if (
                node.output_parameters is not None
                and node.previous_timestamp is not None
                and "platform_params" in node.output_parameters
            ):
                for bus, qubit, parameter, value in node.output_parameters["platform_params"]:
                    parameters[(parameter, bus, qubit)] = (
                        value,
                        node.node_id,
                        datetime.fromtimestamp(node.previous_timestamp),
                    )

        return parameters

    def get_last_fidelities(self) -> dict[tuple, tuple]:
        """Retrieves the last updated fidelities of the graph.

        Returns:
            dict[tuple, tuple]: Fidelities dictionary, with the key and values being tuples containing, in this order:
                - ``key``: (``str``: parameter name, ``int``: qubit).
                - ``value``: (``float``: parameter value, ``str``: node_id where computed, ``datetime``: updated time).
        """
        fidelities: dict[tuple, tuple] = {}
        for node in self.node_sequence.values():
            if (
                node.output_parameters is not None
                and node.previous_timestamp is not None
                and "fidelities" in node.output_parameters
            ):
                for qubit, fidelity, value in node.output_parameters["fidelities"]:
                    fidelities[(fidelity, qubit)] = (
                        value,
                        node.node_id,
                        datetime.fromtimestamp(node.previous_timestamp),
                    )

        return fidelities

    def _dependencies(self, node: CalibrationNode) -> list:
        """Finds the dependencies of a node.

        If in our graph we have `A -> B`, then node B depends on node A. Thus calling this method on B would return A.

        Args:
            node (CalibrationNode): The nodes for which the dependencies need to be retrieved.

        Returns:
            list: The nodes that the argument node depends on.
        """
        return [self.node_sequence[node_name] for node_name in self.calibration_graph.predecessors(node.node_id)]

    @staticmethod
    def _obtain_comparison(node: CalibrationNode, obtained: dict[str, list], comparison: dict[str, list]) -> float:
        """Returns the error, given the chosen method, between the comparison and obtained samples.

        Args:
            node (CalibrationNode): node from which the data comparison will be done.
            obtained (dict): obtained samples to compare.
            comparison (dict): previous samples to compare.

        Returns:
            float: difference/error between the two samples.
        """
        return node.comparison_model(obtained, comparison)

    @staticmethod
    def _is_timeout_expired(timestamp: float, timeout: float) -> bool:
        """Checks if the time passed since the ``timestamp`` is greater than the ``timeout`` duration.

        Args:
            timestamp (float): Timestamp from which the time should be checked, described in UNIX timestamp format.
            timeout (float): The timeout duration in seconds.

        Returns:
            bool: True if the ``timeout`` has expired, False otherwise.
        """
        # Calculate the time that should have passed (timestamp + timeout duration), convert them to datetime objects:
        timestamp_dt = datetime.fromtimestamp(timestamp)
        timeout_duration = timedelta(seconds=timeout)
        timeout_time = timestamp_dt + timeout_duration

        # Check if the current time is greater than the timeout time
        current_time = datetime.now()
        return current_time > timeout_time



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Yokogawa GS200 driver."""
from qcodes.instrument.channel import ChannelTuple, InstrumentModule
from qcodes.instrument_drivers.yokogawa.GS200 import GS200 as QCodesGS200
from qcodes.instrument_drivers.yokogawa.GS200 import GS200_Monitor as QCodesGS200Monitor
from qcodes.instrument_drivers.yokogawa.GS200 import GS200Program as QCodesGS200Program

from qililab.drivers.instruments.instrument_driver_factory import InstrumentDriverFactory
from qililab.drivers.interfaces import BaseInstrument


@InstrumentDriverFactory.register
class GS200(QCodesGS200, BaseInstrument):
    """
    Qililab's driver for the Yokogawa GS200 acting as VoltageSource and CurrentSource

    Args:
        name (str): What this instrument is called locally.
        address (str): The GPIB or USB address of this instrument
        kwargs (Any | dict): kwargs to be passed to VisaInstrument class
    """

    def __init__(self, name: str, address: str, **kwargs):
        super().__init__(name, address, **kwargs)
        self.submodules: dict[str, InstrumentModule | ChannelTuple] = {}  # resetting superclass submodules
        self.instrument_modules: dict[str, InstrumentModule] = {}  # resetting superclass instrument modules
        self._channel_lists: dict[str, ChannelTuple] = {}  # resetting superclass instrument channel lists
        # Add the Monitor to the instrument
        self.add_submodule("measure", QCodesGS200Monitor(self, name="measure", present=True))
        # Add the Program to the instrument
        self.add_submodule("program", QCodesGS200Program(self, name="program"))

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QubitControl class."""
from abc import abstractmethod
from dataclasses import asdict, dataclass
from typing import Sequence

from qpysequence import Sequence as QpySequence

from qililab.constants import RUNCARD
from qililab.instruments.awg_settings.awg_sequencer import AWGSequencer
from qililab.instruments.awg_settings.typings import AWGTypes
from qililab.instruments.instrument import Instrument
from qililab.pulse import PulseBusSchedule
from qililab.utils.asdict_factory import dict_factory


class AWG(Instrument):
    """Abstract base class defining all instruments used to control or readout the qubits."""

    @dataclass(kw_only=True)
    class AWGSettings(Instrument.InstrumentSettings):
        """Contains the settings of a AWG.

        Args:
            num_sequencers (int): Number of sequencers (physical I/Q pairs)
            awg_sequencers (Sequence[AWGSequencer]): Properties of each AWG sequencer
        """

        num_sequencers: int
        awg_sequencers: Sequence[AWGSequencer]

        def __post_init__(self):
            """build AWGSequencers and IQ channels"""
            super().__post_init__()
            if self.num_sequencers <= 0:
                raise ValueError(f"The number of sequencers must be greater than 0. Received: {self.num_sequencers}")
            if len(self.awg_sequencers) != self.num_sequencers:
                raise ValueError(
                    f"The number of sequencers: {self.num_sequencers} does not match"
                    + f" the number of AWG Sequencers settings specified: {len(self.awg_sequencers)}"
                )
            self.awg_sequencers = [
                AWGSequencer(**sequencer) if isinstance(sequencer, dict) else sequencer  # pylint: disable=not-a-mapping
                for sequencer in self.awg_sequencers
            ]

        def to_dict(self):
            """Return a dict representation of an AWG instrument."""
            result = asdict(self, dict_factory=dict_factory)
            result.pop(AWGTypes.AWG_SEQUENCERS.value)

            return result | {AWGTypes.AWG_SEQUENCERS.value: [sequencer.to_dict() for sequencer in self.awg_sequencers]}

    settings: AWGSettings

    @abstractmethod
    def compile(
        self, pulse_bus_schedule: PulseBusSchedule, nshots: int, repetition_duration: int, num_bins: int
    ) -> list:
        """Compiles the ``PulseBusSchedule`` into an assembly program.

        Args:
            pulse_bus_schedule (PulseBusSchedule): the list of pulses to be converted into a program
            nshots (int): number of shots / hardware average
            repetition_duration (int): repetition duration
            num_bins (int): number of bins

        Returns:
            list: list of compiled assembly programs
        """

    @abstractmethod
    def run(self, port: str):
        """Run the uploaded program"""

    @abstractmethod
    def upload_qpysequence(self, qpysequence: QpySequence, port: str):
        """Upload qpysequence."""

    @abstractmethod
    def upload(self, port: str):
        """Upload compiled program."""

    @property
    def num_sequencers(self):
        """Number of sequencers in the AWG

        Returns:
            int: number of sequencers
        """
        return self.settings.num_sequencers

    @property
    def awg_sequencers(self):
        """AWG 'awg_sequencers' property."""
        return self.settings.awg_sequencers

    @property
    def intermediate_frequencies(self):
        """AWG 'intermediate_frequencies' property."""
        return [sequencer.intermediate_frequency for sequencer in self.awg_sequencers]

    def to_dict(self):
        """Return a dict representation of an AWG instrument."""
        return {RUNCARD.NAME: self.name.value} | self.settings.to_dict()

    def get_sequencers_from_chip_port_id(self, chip_port_id: str):
        """Get sequencer ids from the chip port identifier

        Args:
            chip_port_id (str): chip port identifier

        Returns:
            list[AWGSequencer]: list of integers containing the indices of the sequencers connected to the chip port
        """
        if seqs := [sequencer for sequencer in self.awg_sequencers if sequencer.chip_port_id == chip_port_id]:
            return seqs
        raise IndexError(
            f"No sequencer found connected to port {chip_port_id}. Please make sure the `chip_port_id` "
            "attribute is correct."
        )

    def get_sequencer(self, sequencer_id: int) -> AWGSequencer:
        """Get sequencer from the sequencer identifier

        Args:
            sequencer_id (int): sequencer identifier

        Returns:
            AWGSequencer: sequencer associated with the sequencer_id
        """
        sequencer_identifiers = [
            sequencer.identifier for sequencer in self.awg_sequencers if sequencer.identifier == sequencer_id
        ]

        if len(sequencer_identifiers) != 1:
            raise ValueError(
                f"Each sequencer should have a unique id. Found {len(sequencer_identifiers)} sequencers "
                f"with id {sequencer_id}."
            )

        return self.awg_sequencers[sequencer_identifiers[0]]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QuantumMachinesResult class."""
import numpy as np

from qililab.constants import QMRESULT, RUNCARD
from qililab.result.counts import Counts
from qililab.result.result import Result
from qililab.typings.enums import ResultName
from qililab.utils.factory import Factory


@Factory.register
class QuantumMachinesMeasurementResult(Result):
    """Contains the data obtained from a single measurment in Quantum Machines hardware.

    Args:
        I (np.ndarray): Data obtained from I stream.
        Q (np.ndarray, Optional): Data obtained from Q stream. Defaults to None.
        adc1 (np.ndarray, Optional): Data obtained from adc1 stream. Defaults to None.
        adc2 (np.ndarray, Optional): Data obtained from adc2 stream. Defaults to None.
    """

    name = ResultName.QUANTUM_MACHINES_MEASUREMENT

    def __init__(
        self, I: np.ndarray, Q: np.ndarray | None = None, adc1: np.ndarray | None = None, adc2: np.ndarray | None = None
    ):
        self.I = I
        self.Q = Q
        self.adc1 = adc1
        self.adc2 = adc2

    @property
    def array(self) -> np.ndarray:
        """Returns I/Q data as a combined numpy array.

        Returns:
            np.ndarray: The I/Q data as a compined nummpy array.
        """

        return (
            np.concatenate((self.I.reshape(1, *self.I.shape), self.Q.reshape(1, *self.Q.shape)), axis=0)
            if self.Q is not None
            else self.I.reshape(1, *self.I.shape)
        )

    def to_dict(self) -> dict:
        """Returns a serialized dictionary of the QuantumMachinesResult class.

        Returns:
            dict[str: str | np.ndarray]: Dictionary containing all the class information.
        """
        return {
            RUNCARD.NAME: self.name.value,
            QMRESULT.I: self.I,
            QMRESULT.Q: self.Q,
            QMRESULT.ADC1: self.adc1,
            QMRESULT.ADC2: self.adc2,
        }

    def probabilities(self) -> dict[str, float]:
        """Return probabilities of being in the ground and excited state.

        Returns:
            dict[str, float]: Dictionary containing the quantum states as the keys of the dictionary, and the
                probabilities obtained for each state as the values of the dictionary.
        """
        raise NotImplementedError("Probabilities are not yet supported for Quantum Machines instruments.")

    def counts_object(self) -> Counts:
        """Returns a Counts object containing the amount of times each state was measured.

        Raises:
            NotImplementedError: Not implemented.

        Returns:
            Counts: Counts object containing the amount of times each state was measured.
        """
        raise NotImplementedError("Counts are not yet supported for Quantum Machines instruments.")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantum Machines OPX class."""
from dataclasses import dataclass
from typing import Dict

import numpy as np
from qm import QuantumMachine
from qm import QuantumMachinesManager as QMM
from qm import SimulationConfig
from qm.jobs.running_qm_job import RunningQmJob
from qm.qua import Program

from qililab.instruments.instrument import Instrument
from qililab.instruments.utils import InstrumentFactory
from qililab.typings import InstrumentName, Parameter, QMMDriver


@InstrumentFactory.register
class QuantumMachinesManager(Instrument):
    """Class defining the Qililab Quantum Machines Manager instrument in Qililab.

    This class allows Qililab control and communication with an instance of the
    Quantum Machines Manager, which is the central class to interact with Quantum Machines instruments.
    The manager is responsible, through the use of QUA sequences, of setting the Quantum Machines instruments
    in the right manner for Quantum Control.

    Args:
        name (InstrumentName): Name of the Instrument.
        device (QMMDriver): Instance of the Quantum Machines Manager Driver class.
        settings (QMMSettings): Settings of the instrument.
    """

    name = InstrumentName.QUANTUM_MACHINES_MANAGER

    @dataclass
    class QMMSettings(Instrument.InstrumentSettings):
        """Settings for Quantum Machines Manager instrument.

        Args:
            qop_ip (str): I.P. address to connect to the Quantum Machines instruments.
            qop_port (int): Port to connect to the Quantum Machines instruments.
            config (Dict): Configuration dictionary for the Quantum Machines instruments.
        """

        qop_ip: str
        qop_port: int
        config: Dict

    settings: QMMSettings
    device: QMMDriver
    qm: QuantumMachine

    @Instrument.CheckDeviceInitialized
    def initial_setup(self):
        """Sets initial instrument settings.

        Creates an instance of the Qilililab Quantum Machines Manager, and the Quantum Machine, opening
        a connection to the Quantum Machine by the use of the Quantum Machines Manager.
        """
        super().initial_setup()
        qmm = QMM(host=self.settings.qop_ip, port=self.settings.qop_port)
        self.qm = qmm.open_qm(config=self.settings.config, close_other_machines=True)

    @Instrument.CheckDeviceInitialized
    def turn_on(self):
        """Turns on an instrument."""

    @Instrument.CheckDeviceInitialized
    def reset(self):
        """Resets instrument settings."""

    @Instrument.CheckDeviceInitialized
    def turn_off(self):
        """Turns off an instrument."""

    def set_parameter(self, parameter: Parameter, value: float | str | bool, channel_id: int | None = None):
        """Sets the parameter of a specific instrument.

        Args:
            parameter (Parameter): parameter settings of the instrument to update
            value (float | str | bool): value to update
            channel_id (int | None, optional): instrument channel to update, if multiple. Defaults to None.

        Returns:
            bool: True if the parameter is set correctly, False otherwise
        """
        raise NotImplementedError("Setting a parameter is not supported for Quantum Machines yet.")

    def run(self, program: Program) -> RunningQmJob:
        """Runs the QUA Program.

        Creates, for every execution, an instance of a QuantumMachine class and executes the QUA program on it.
        After execution, the job generated by the Quantum Machines Manager is returned.

        Args:
            program (Program): QUA Program to be run on Quantum Machines instruments.

        Returns:
            job: Quantum Machines job.
        """

        return self.qm.execute(program)

    def get_acquisitions(self, job: RunningQmJob) -> dict[str, np.ndarray]:
        """Fetches the results from the execution of a QUA Program.

        Once the results have been fetched, they are returned wrapped in a QuantumMachinesResult instance.

        Args:
            job (QmJob): Job that provides the result handles.

        Returns:
            QuantumMachinesResult: Quantum Machines result instance.
        """
        result_handles_fetchers = job.result_handles
        result_handles_fetchers.wait_for_all_values()
        results = {
            name: handle.fetch_all(flat_struct=True) for name, handle in job.result_handles if handle is not None
        }
        return {name: data for name, data in results.items() if data is not None}

    def simulate(self, program: Program) -> RunningQmJob:
        """Simulates the QUA Program.

        Creates, for every simulation, an instance of a QuantumMachine class and executes the QUA program on it.
        It returns the running job instance.

        Args:
            program (Program): QUA Program to be simulated on Quantum Machines instruments.

        Returns:
            job: Quantum Machines job.
        """
        return self.qm.simulate(program, SimulationConfig(40_000))



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ReadoutSystemControl class."""
from qililab.instruments import AWGAnalogDigitalConverter
from qililab.result import Result
from qililab.typings.enums import SystemControlName
from qililab.utils import Factory

from .system_control import SystemControl


@Factory.register
class ReadoutSystemControl(SystemControl):
    """System control used for readout."""

    name = SystemControlName.READOUT_SYSTEM_CONTROL

    def acquire_result(self) -> Result:
        """Read the result from the vector network analyzer instrument

        Returns:
            Result: Acquired result
        """
        # TODO: Support acquisition from multiple instruments
        results: list[Result] = []
        for instrument in self.instruments:
            result = instrument.acquire_result()
            if result is not None:
                results.append(result)

        if len(results) > 1:
            raise ValueError(
                f"Acquisition from multiple instruments is not supported. Obtained a total of {len(results)} results. "
            )

        return results[0]

    def acquire_qprogram_results(self, acquisitions: list[str]) -> list[Result]:
        """Read the result from the vector network analyzer instrument

        Returns:
            list[Result]: Acquired results in chronological order
        """
        # TODO: Support acquisition from multiple instruments
        total_results: list[list[Result]] = []
        for instrument in self.instruments:
            instrument_results = instrument.acquire_qprogram_results(acquisitions=acquisitions)
            total_results.append(instrument_results)

        return total_results[0]

    @property
    def acquisition_delay_time(self) -> int:
        """SystemControl 'acquisition_delay_time' property.
        Delay (in ns) between the readout pulse and the acquisition."""
        for instrument in self.instruments:
            if isinstance(instrument, AWGAnalogDigitalConverter):
                return instrument.acquisition_delay_time
        raise ValueError(f"The system control {self.name.value} doesn't have an AWG instrument.")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Automatic-calibration Node module, which works with notebooks as nodes."""
import json
import logging
import os
from datetime import datetime
from io import StringIO
from typing import Callable

import numpy as np
import papermill as pm

from qililab.config import logger

logger_output_start = "RAND_INT:47102512880765720413 - OUTPUTS: "


class CalibrationNode:  # pylint: disable=too-many-instance-attributes
    """Automatic calibration Node class representing a node in the calibration graph.

    The calibration graph represents a calibration procedure, where each node represents a step of this calibration procedure. **Each of these steps consists of:**

    - **A Jupyter Notebook** to calibrate or check the data with (including its metadata for execution time checks). Such a notebook should contain the following:

        **1) An input parameters cell** (tagged as `parameters`). These parameters are the ones to be overwritten by the ``input_parameters``.

        **2) An experiment/circuit** with its corresponding loops for the sweep given by ``sweep_interval``.

        **3) An analysis procedure**, that plots and fits the obtained data to the expected theoretical behavior, finding the optimal desired parameters.

        **4) An export data cell**, that calls ``export_nb_outputs()`` with the dictionary to retrieve data from the notebook into the calibration workflow.
        This dictionary contains a ``check_parameters`` dictionary of the obtained results for comparisons and a ``platform_params`` list of parameters to set on the platform.

        .. note::

            More information about the notebooks contents and execution, can be found in the examples below.

    - **Thresholds and Models for the Comparison** of the data and metadata of this notebook, such:
        ``in_spec_threshold``, ``bad_data_threshold``, ``comparison_model``, ``drift_timeout``.

    - **Inputs to pass to this notebook (optional)**, which might vary for different calls of the same notebook, such:
        ``sweep_interval``, ``number_of_random_datapoints``, ``input_parameters`` (kwargs).

    .. note::

        If the same notebook is going to be used multiple times for the same ``qubit``, then you need to pass a ``node_distinguisher`` argument, so the :class:`.CalibrationController`
        makes the graph mapping correctly.

    Args:
        nb_path (str): Full notebook path with the folder, nb_name, and ``.ipynb`` extension, written in unix format: `folder/subfolder/.../file.ipynb`.
        in_spec_threshold (float): Threshold such that the ``check_data()`` methods return `in_spec` or `out_of_spec`.
        bad_data_threshold (float): Threshold such that the ``check_data()`` methods return `out_of_spec` or `bad_data`.
        comparison_model (Callable): Comparison model used, to compare data in this node.
        drift_timeout (float): Duration in seconds, representing an estimate of how long it takes for the parameter to drift. During that time the parameters of
            this node should be considered calibrated without the need to check the data.
        qubit_index (int | list[int] | None, optional): Qubit on which this notebook will be executed. Defaults to None.
        node_distinguishier (int | str | None, optional): Distinguisher for when the same notebook its used multiple times in the same qubit. Mandatory to use in such case, or
            the :class:`.CalibrationController` won't do the graph mapping properly, and the calibration will fail. Defaults to None.
        input_parameters (dict | None, optional): Kwargs for input parameters to pass and be interpreted by the notebook. Defaults to None.
        sweep_interval (np.ndarray | None, optional): Array describing the sweep values of the experiment. Defaults to None, which means the one specified in the notebook will be used.
        number_of_random_datapoints (int, optional): The number of points randomly choose within the sweep interval, to run ``check_data()`` with. Default value is 10.

    Examples:

        **Notebook execution:**

        First the key functionality of this class is implemented in the ``run_node()`` method**. The workflow of ``run_node()`` is as follows:

        1. Prepare any input parameters needed for the notebook, including extra parameters defined by the user and essential ones such as the targeted qubit or the sweep intervals.

        2. Create a file with a temporary name. This file will be used to save the execution of the notebook and initially has the following format:

            ``NameOfTheNode_TimeExecutionStarted_dirty.ipynb``

            The ``_dirty`` flag is added to identify executions that are not completed. Since the data we would find such file is ``dirty``, not completed.

        3. Start the execution of the notebook. There are three possible outcomes:

            3.1) The execution succeeds. If the execution succeeds, the execution file is renamed by updating the timestamp and removing the dirty flag:

                ``NameOfTheNode_TimeExecutionEnded.ipynb``

            3.2) The execution is interrupted. If the execution is interrupted, the ``_dirty`` flag remains in the filename, and the program exits:

                ``NameOfTheNode_TimeExecutionStarted_dirty.ipynb``

            3.2) An exception is thrown. This case is not controlled by the user like interruptions. Instead, exceptions are automatically thrown when
            an error is detected. When an execution error is found, the execution file is moved to a new subfolder ``/error_executions`` and renamed with the
            time the error occurred, adding the `_error` flag, and the program exits:

                ``NameOfTheNode_TimeExecutionFoundError_error.ipynb``

            A more detailed explanation of the error is reported and also described inside the notebook (see `papermill documentation
            <https://papermill.readthedocs.io/en/latest/>`_ for more detailed information).

        At the end of this process, you obtain an executed and saved notebook for manual inspection, along with the optimal parameters to set in the runcard
        and the achieved fidelities.

        ----------

        **Practical example:**

        To create two linked nodes, and pass them to a :class:`.CalibrationController`**, you need:

        .. code-block:: python

            import numpy as np
            import networkx as nx

            from qililab.calibration import CalibrationController, CalibrationNode, norm_root_mean_sqrt_error

            # GRAPH CREATION AND NODE MAPPING (key = name in graph, value = node object):
            nodes = {}
            G = nx.DiGraph()
            qubit = 0

            # CREATE NODES :
            for qubit in [0, 1]:
                first = CalibrationNode(
                    nb_path="notebooks/first.ipynb",
                    qubit_index=qubit,
                    in_spec_threshold=4,
                    bad_data_threshold=8,
                    comparison_model=norm_root_mean_sqrt_error,
                    drift_timeout=1800.0,
                )
                nodes[first.node_id] = first

                second = CalibrationNode(
                    nb_path="notebooks/second.ipynb",
                    qubit_index=qubit,
                    in_spec_threshold=2,
                    bad_data_threshold=4,
                    comparison_model=norm_root_mean_sqrt_error,
                    drift_timeout=1.0,
                    sweep_interval=np.arange(start=0, stop=19, step=1),
                )
                nodes[second.node_id] = second

                # GRAPH BUILDING (1 --> 2):
                G.add_edge(first.node_id, second.node_id)

            # CREATE CALIBRATION CONTROLLER:
            controller = CalibrationController(node_sequence=nodes, calibration_graph=G, runcard=path_runcard)

            ### WORKFLOW TO DO:
            controller.maintain(nodes["second_q1"]) # maintain second node for qubit 1

        .. note::

            You can find the above code, but defining ``first`` and ``second`` as lists, in the :class:`CalibrationController` class documentation.


        |

        where **the notebooks ``first/second.ipynb``, would contain** the following:

        |

        **1) An input parameters cell** (tagged as `parameters`). These parameters are the ones to be overwritten by the ``input_parameters``:

            .. code-block:: python

                import numpy as np

                # ``check_data()`` parameters:
                check = False
                number_of_random_datapoints = 10
                qubit=0

                # Sweep interval:
                sweep_interval = np.arange(start=0, stop=19, step=1)

                # Extra parameters for this concrete notebook:
                param1=0
                param2=0
                ...

        |

        **2) An experiment/circuit** with its corresponding loops for the sweep given by ``sweep_interval``.

            .. code-block:: python

                # Set the environment and paths:
                ...

                # Set the platform:
                platform = ql.build_platform(path=platform_path)
                platform.connect()
                platform.initial_setup()
                platform.turn_on_instruments()

                # Define circuit
                circuit = ...

                # Loop over the sweeps executing the platform:
                results_list = []
                for X in sweep_interval:
                    platform.set_parameter(alias=alias, parameter=ql.Parameter.X, value=X)
                    result = platform.execute(program=circuit, num_avg=hw_avg, repetition_duration=repetition_duration)
                    results_list.append(result.array)

                results = np.hstack(results_list)

        |

        **3) An analysis procedure**, that plots and fits the obtained data to the expected theoretical behaviour and finds the optimal desired parameters.

            .. code-block:: python

                def fit(xdata, results):
                    ...

                fitted_values, x_data, y_data, figure = fit(xdata=sweep_interval, results=results)
                plt.show()

        |

        **4) An export data cell**, that calls ``export_nb_outputs()`` with the dictionary to retrieve from the notebook into the calibration workflow:

            .. code-block:: python

                from qililab.automatic_calibration.calibration_node import export_nb_outputs

                export_nb_outputs(
                    {
                        "check_parameters": {"x": sweep_interval, "y": results},
                        "platform_params": [(bus_alias0, qubit, param_name0, fitted_values[0]), (bus_alias1, qubit, param_name1, fitted_values[1])],
                        "fidelities": [(qubit, "fidelity1", 0.9), (qubit, "fidelity2", 0.95)]  # Fidelities in the output dictionary are optional.
                    }
                )

        where the ``check_parameters`` are a dictionary of the saved results to do comparisons against. And the ``platform_params`` are a list of parameters to set on the platform.
    """

    def __init__(
        self,
        nb_path: str,
        in_spec_threshold: float,
        bad_data_threshold: float,
        comparison_model: Callable,
        drift_timeout: float,
        qubit_index: int | list[int] | None = None,
        node_distinguisher: int | str | None = None,
        input_parameters: dict | None = None,
        sweep_interval: np.ndarray | None = None,
        number_of_random_datapoints: int = 10,
    ):
        if in_spec_threshold > bad_data_threshold:
            raise ValueError("`in_spec_threshold` must be smaller or equal than `bad_data_threshold`.")

        if len(nb_path.split("\\")) > 1:
            raise ValueError("`nb_path` must be written in unix format: `folder/subfolder/.../file.ipynb`.")

        self.nb_path: str = nb_path
        """Absolute notebook path, with folder, nb_name and ``.ipynb`` extension."""

        self.qubit_index: int | list[int] | None = qubit_index
        """Qubit which this notebook will be executed on."""

        self.node_distinguisher: int | str | None = node_distinguisher
        """Distinguisher for when the same notebook its used multiple times in the same qubit. Mandatory to use in such case, or
        the :class:`.CalibrationController` won't do the graph mapping properly, and the calibration will fail. Defaults to None.
        """

        self.node_id, self.nb_folder = self._path_to_name_and_folder(nb_path)
        """Node name and folder, separated, and without the ``.ipynb`` extension."""

        self.in_spec_threshold: float = in_spec_threshold
        """Threshold such that the ``check_data()`` methods return `in_spec` or `out_of_spec`."""

        self.bad_data_threshold: float = bad_data_threshold
        """Threshold such that the ``check_data()`` methods return `out_of_spec` or `bad_data`."""

        self.comparison_model: Callable = comparison_model
        """Comparison model used, to compare data in this node."""

        self.drift_timeout: float = drift_timeout
        """A durations in seconds, representing an estimate of how long it takes for the parameter to drift. During that time the parameters of
        this node should be considered calibrated, without the need to check the data.
        """

        self.input_parameters: dict | None = input_parameters
        """Kwargs for input parameters, to pass and then be interpreted by the notebook. Defaults to None."""

        self.sweep_interval: np.ndarray | None = sweep_interval
        """Array describing the sweep values of the experiment. Defaults to None, which means the one specified in the notebook will be used."""

        self.number_of_random_datapoints: int = number_of_random_datapoints
        """The number of points, chosen randomly within the sweep interval, to check with ``check_data()`` if the experiment
        gets the same outcome as during the last calibration that was run. Default value is 10.
        """

        self.output_parameters: dict | None = self.get_last_calibrated_output_parameters()
        """Output parameters dictionary from the notebook execution, which was extracted with ``ql.export_nb_outputs()``, normally contains
        a ``check_params`` to do the ``check_data()`` and the ``platform_params`` which will be the calibrated parameters to set in the platform.

        If no previous successful calibration, then is None.
        """

        self.previous_output_parameters: dict | None = None
        """Same output_parameters, but from the previous execution of the Node. Starts at None."""

        self.previous_timestamp: float | None = self.get_last_calibrated_timestamp()
        """Last calibrated timestamp. If no previous successful calibration, then is None."""

        self._stream: StringIO = self._build_notebooks_logger_stream()
        """Stream object to which the notebooks logger output will be written, to posterior retrieval."""

    def run_node(self, check: bool = False) -> float:
        """Executes the notebook, passing the needed parameters and flags. Also it can be chosen to only check certain values of the sweep interval for
        when checking data.

        **Its workflow is the following:**

        1. Prepare any input parameters needed for the notebook. This includes extra parameters defined by the user and essential ones such as the targeted qubit or the sweep intervals.

        2. Create a file with a temporary name. This file will be used to save the execution of the notebook, and at this stage has the following format:

            ``NameOfTheNode_TimeExecutionStarted_dirty.ipynb``

            The "_dirty" flag is added to the execution files in order to identify the executions that are not completed. So we know that the data we find
            if we open that file is "dirty", not completed.

        3. Start execution of the notebook. From where we can expect 3 possible outcomes from the execution:

            3.1) The execution succeds. If the execution succeds, we rename the execution file by updating the timestamp and removing the dirty flag:

                ``NameOfTheNode_TimeExecutionEnded.ipynb``

            3.2) The execution is interrupted. If the execution is interrupted, the "_dirty" flag will remain in the filename for ever. Notice after an
            interruption the program exits.

                ``NameOfTheNode_TimeExecutionStarted_dirty.ipynb``

            3.2) An exception is thrown. This case is not controlled by the user like the interruptions, instead those exceptions are automatically
            thrown when an error is detected. When an execution error is found, the execution file is moved to a new subfolder ``/error_executions``
            and renamed with the time that the error ocurred and adding the flag "_error":

                ``NameOfTheNode_TimeExecutionFoundError_error.ipynb``

            A more detailed explanation of the error is reported and also described inside the notebook (see `papermill documentation
            <https://papermill.readthedocs.io/en/latest/>`_ for more detailed information).Then after we post-processed the file, the program exits.

        at the end, from all of this, you will obtain, a executed and saved notebook to manually check, and an outputs dictionary, containing the optimal
        parameters to set in the runcard, together with the achieved fidelities, and the data for future comparisons.

        Args:
            check (bool, optional): If True, runs a ``check_data()`` (randomly ``number_of_datapoints`` selected points from ``sweep_interval``) instead
            than a normal full length ``calibrate()``. Defaults to ``False``.

        Returns:
            float: Timestamp to identify the notebook execution moment.

        Exits:
            In case of a keyboard interruption or any exception during the execution of the notebook.
        """
        # Create the input parameters for the notebook:)
        params: dict = {
            "check": check,
            "number_of_random_datapoints": self.number_of_random_datapoints,
            "qubit": self.qubit_index,
        }

        if self.sweep_interval is not None:
            params["sweep_interval"] = self._build_check_data_interval() if check else self.sweep_interval

        if self.input_parameters is not None:
            params |= self.input_parameters

        # initially the file is "dirty" until we make sure the execution was not aborted
        output_path = self._create_notebook_datetime_path(self.nb_path, dirty=True)
        self.previous_output_parameters = self.output_parameters

        # Execute notebook without problems:
        try:
            self.output_parameters = self._execute_notebook(self.nb_path, output_path, params)

            timestamp = datetime.timestamp(datetime.now())
            new_output_path = self._create_notebook_datetime_path(self.nb_path, timestamp)
            os.rename(output_path, new_output_path)
            return timestamp

        # When keyboard interrupt (Ctrl+C), generate error, and leave `_dirty`` in the name:
        except KeyboardInterrupt as exc:
            logger.error("Interrupted automatic calibration notebook execution of %s", self.nb_path)
            raise KeyboardInterrupt(f"Interrupted automatic calibration notebook execution of {self.nb_path}") from exc

        # When notebook execution fails, generate error folder and move there the notebook:
        except Exception as exc:  # pylint: disable = broad-exception-caught
            timestamp = datetime.timestamp(datetime.now())
            error_path = self._create_notebook_datetime_path(self.nb_path, timestamp, error=True)
            os.rename(output_path, error_path)
            logger.error(
                "Aborting execution. Exception %s during automatic calibration notebook execution, trace of the error can be found in %s",
                str(exc),
                error_path,
            )
            # pylint: disable = broad-exception-raised
            raise Exception(
                f"Aborting execution. Exception {str(exc)} during automatic calibration notebook execution, trace of the error can be found in {error_path}"
            ) from exc

    @staticmethod
    def _build_notebooks_logger_stream() -> StringIO:
        """Build a stream object for capturing the notebook's log output.

        Returns:
            StringIO: Stream object for capturing execution outputs.
        """
        stream = StringIO()
        logging.basicConfig(stream=stream, level=logging.INFO)

        return stream

    def _execute_notebook(self, input_path: str, output_path: str, parameters: dict | None = None) -> dict:
        """Executes a Jupyter Notebook overwriting the `parameters` cell, and capturing the execution ``output``.

        Args:
            input_path (str): The input path of the notebook to be executed.
            output_path (str): The output path where the executed noteboo will be saved. If None, no file will be saved.
            parameters (dict | None, optional): Input parameters kwargs, to overwrite the notebook `parameters` cell with. Defaults to None.

        Returns:
            dict: A dictionary containing the output parameters of the execution.

        Raises:
            IncorrectCalibrationOutput: In case no outputs, incorrect outputs or multiple outputs where found. Incorrect outputs are those that do not contain `check_parameters` or is empty.
        """
        pm.execute_notebook(input_path, output_path, parameters, log_output=True, stdout_file=self._stream)

        # Retrieve the logger info and extract the output from it:
        logger_string = self._stream.getvalue()
        return self._from_logger_string_to_output_dict(logger_string, input_path)

    def _build_check_data_interval(self) -> np.ndarray | None:
        """Builds ``check_data()`` sweep interval with ``number_of_random_datapoints`` data points.

        Returns:
            np.ndarray | None: An array representing the sweep interval for checking data, or None if not specified.
        """
        if (interval := self.sweep_interval) is not None:
            return np.array(
                [interval[np.random.randint(0, len(interval))] for _ in range(self.number_of_random_datapoints)]
            )
        return None

    def _create_notebook_datetime_path(
        self, original_path: str, timestamp: float | None = None, dirty: bool = False, error: bool = False
    ) -> str:
        """Create a timestamped notebook path, adding the datetime to the file name end, just before the ``.ipynb``.

        If the path directory doesn't exist, it gets created.

        The passed path can have the ``.ipynb`` extension or not.

        The part of the string after the last "/" will be considered the file name, and the part before its directory.

        Args:
            original_path (str): The original path of the notebook, to add the datetime to. The path directory doesn't need to exist. Can have the ``.ipynb`` extension or not.
                The part of the string after the last "/" will be considered the file name, and the part before it's directory.
            timestamp (float | None, optional): Timestamp to add to the name. If None, the current time will be used. Defaults to None.
            dirty (bool, optional): Flag indicating if the notebook is in a "dirty" state. Defaults to False.
            error (bool, optional): Flag indicating if the notebook comes from an execution error. Defaults to False.

        Returns:
            str: The timestamped notebook path.
        """
        # Create datetime pathHM
        now = datetime.now() if timestamp is None else datetime.fromtimestamp(timestamp)
        daily_path = f"{now.year}_{now.month:02d}_{now.day:02d}"
        now_path = f"{daily_path}-" + f"{now.hour:02d}:{now.minute:02d}:{now.second:02d}"

        # If doesn't exist, create the needed folder for the path
        name, folder_path = self._path_to_name_and_folder(original_path)
        os.makedirs(folder_path, exist_ok=True)

        if dirty and not error:  # return the path of the execution
            return f"{folder_path}/{name}_{now_path}_dirty.ipynb"
        if error:
            os.makedirs(f"{folder_path}/error_executions", exist_ok=True)
            return f"{folder_path}/error_executions/{name}_{now_path}_error.ipynb"

        # return the string where saved
        return f"{folder_path}/{name}_{now_path}.ipynb"

    def _path_to_name_and_folder(self, original_path: str) -> tuple[str, str]:
        """Extract the name and folder from a notebook path. Name will be extended with the qubit it acts on.

        The passed path can have the ``.ipynb`` extension or not.

        The part of the string after the last "/" will be considered the file name, and the part before its directory.

        Args:
            original_path (str): The original path of the notebook. Can have the ``.ipynb`` extension or not.
                The part of the string after the last "/" will be considered the file name, and the part before it's directory.

        Returns:
            tuple[str, str]: A tuple containing the notebook name and its folder.
        """
        # Create qubit_string to add:
        qubit_str = (
            f"_q{str(self.qubit_index)}"
            if isinstance(self.qubit_index, int)
            else "_" + "".join(f"q{q}" for q in self.qubit_index)
            if isinstance(self.qubit_index, list)
            else ""
        )

        # Create distinguish_string to differentiate multiple calls of the same node:
        distinguish_str = f"_{str(self.node_distinguisher)}" if self.node_distinguisher is not None else ""

        # Remove .ipynb from end if it has one, and separate the folder and name with the last "/":
        path_list = original_path.split(".ipynb")[0].split("/")

        name = path_list.pop() + distinguish_str + qubit_str
        folder_path = "/".join(path_list)
        return name, folder_path

    def get_last_calibrated_timestamp(self) -> float | None:
        """Gets the timestamp of the last successful calibration of the notebook.

        Searches the directory for self.node_id+“_calibrated” and gets the latest creation time.

        Returns:
            float | None: The timestamp of the last successful calibration, or None if no calibration has been completed.
        """
        last_modified_file_name = self._find_last_executed_calibration()
        return (
            os.path.getmtime(os.path.join(self.nb_folder, last_modified_file_name))
            if last_modified_file_name is not None
            else None
        )

    def get_last_calibrated_output_parameters(self) -> dict | None:
        """Gets the output parameters of the last successful calibration.

        Searches the directory for self.node_id+“_calibrated” and gets the outputs from the latest creation time one.

        Returns:
            dict | None: The output parameters of the last successful calibration, or None if no calibration has been completed.
        """
        last_modified_file_name = self._find_last_executed_calibration()
        return (
            self._parse_output_from_execution_file(last_modified_file_name)
            if last_modified_file_name is not None
            else None
        )

    def _find_last_executed_calibration(self) -> str | None:
        """Finds the last executed calibration of the node.

        Returns:
            str | None: The path of the last calibrated notebook, or None if none has been executed.
        """
        # Filtering by node_id and calibrations
        file_names = [
            entry.name
            for entry in os.scandir(self.nb_folder)
            if entry.is_file()
            and f"{self.node_id}" in entry.name
            and "_calibrated" in entry.name.split(f"{self.node_id}")[1]
        ]
        # Get the last created file, most recent one
        return (
            max(file_names, key=lambda fname: os.path.getctime(os.path.join(self.nb_folder, fname)))
            if file_names != []
            else None
        )

    def _parse_output_from_execution_file(self, file_name: str) -> dict | None:
        """Parses the output information, from a notebook execution file.

        Args:
            file_name (str): The name of the execution file to parse.

        Returns:
            dict | None: A dictionary containing parsed output information or None if parsing fails.

        Raises:
            IncorrectCalibrationOutput: In case no outputs, incorrect outputs or multiple outputs where found. Incorrect outputs are those that do not contain `check_parameters` or is empty.
        """
        # Parsing file
        outputs_lines: list[str] = []
        try:
            with open(os.path.join(self.nb_folder, file_name), encoding="utf-8") as file:
                lines = file.readlines()
                outputs_lines.extend(line for line in lines if line.find(logger_output_start) != -1)
        except Exception as exc:
            logger.error("No previous execution found of notebook %s.", self.nb_path)
            raise FileNotFoundError(f"No previous execution found of notebook {self.nb_path}.") from exc

        # Check how many lines contain an output, to raise the corresponding errors:
        if len(outputs_lines) != 1:
            logger.error("No output or various outputs found in notebook %s.", self.nb_path)
            raise IncorrectCalibrationOutput(f"No output or various outputs found in notebook {self.nb_path}.")

        # When only one line of outputs, use that one:
        return self._from_logger_string_to_output_dict(outputs_lines[0], self.nb_path)

    def _from_logger_string_to_output_dict(self, logger_string: str, input_path: str) -> dict:
        """Returns the output dictionary from a logger output string. Raises errors if the ouput doesn't follow the expected format.

        Args:
            logger_string (str): The logger string containing the output dictionary to extract.
            input_path (str): Path of the notebook that generated the output, to raise errors.

        Returns:
            dict: The output dictionary of the file execution.

        Raises:
            IncorrectCalibrationOutput: In case no outputs, incorrect outputs or multiple outputs where found. Incorrect outputs are those that do not contain `check_parameters` or is empty.
        """
        logger_splitted = logger_string.split(logger_output_start)

        # In case something unexpected happened with the output we raise an error
        if len(logger_splitted) != 2:
            logger.error("No output or various outputs found in notebook %s.", input_path)
            raise IncorrectCalibrationOutput(f"No output or various outputs found in notebook {input_path}.")

        # This next line is for taking into account other encodings, where special characters get `\\` in front.
        clean_data = logger_splitted[1].split("\\n")[0].replace('\\"', '"')

        logger_outputs_string = clean_data.split("\n")[0]
        out_dict = json.loads(logger_outputs_string)

        if "check_parameters" not in out_dict or out_dict["check_parameters"] == {}:
            logger.error(
                "Aborting execution. No 'check_parameters' dictionary or its empty in the output cell implemented in %s",
                input_path,
            )
            raise IncorrectCalibrationOutput(
                f"Empty output found in {input_path}, output must have key and value 'check_parameters'."
            )
        return out_dict

    def _add_string_to_checked_nb_name(self, string_to_add: str, timestamp: float) -> None:
        """Adds a string to the checked notebook name.

        Args:
            string_to_add (str): The string to append to the notebook name.
            timestamp (float): The timestamp to use for the new notebook execution name.
        """
        path = os.path.join(self.nb_folder, self.node_id)
        timestamp_path = self._create_notebook_datetime_path(path, timestamp).split(".ipynb")[0]

        os.rename(f"{timestamp_path}.ipynb", f"{timestamp_path}_{string_to_add}.ipynb")


def export_nb_outputs(outputs: dict) -> None:
    """Function to export notebook outputs into a stream, later collected by the :class:`CalibrationNode` class.

    Args:
        outputs (dict): Outputs from the notebook to export into the automatic calibration workflow.
    """
    print(f"{logger_output_start}{json.dumps(outputs)}")


class IncorrectCalibrationOutput(Exception):
    """Error raised when the output of a calibration node is incorrect."""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"IncorrectCalibrationOutput: {self.message}"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SystemControl class."""

import contextlib
from abc import ABC
from dataclasses import InitVar, dataclass
from typing import get_type_hints

from qpysequence import Sequence as QpySequence

from qililab.constants import RUNCARD
from qililab.instruments import AWG, Instrument, Instruments
from qililab.instruments.instrument import ParameterNotFound
from qililab.instruments.qblox import QbloxModule
from qililab.pulse import PulseBusSchedule
from qililab.settings import Settings
from qililab.typings import FactoryElement
from qililab.typings.enums import Parameter, SystemControlName
from qililab.utils import Factory


@Factory.register
class SystemControl(FactoryElement, ABC):
    """SystemControl class."""

    name = SystemControlName.SYSTEM_CONTROL

    @dataclass(kw_only=True)
    class SystemControlSettings(Settings):
        """SystemControlSettings class."""

        instruments: list[Instrument]
        platform_instruments: InitVar[Instruments]

        def __post_init__(self, platform_instruments: Instruments):  # type: ignore # pylint: disable=arguments-differ
            # ``self.instruments`` contains a list of instrument aliases
            instruments = []
            for inst_alias in self.instruments:
                inst_class = platform_instruments.get_instrument(alias=inst_alias)  # type: ignore
                if inst_class is None:
                    raise NameError(
                        f"The instrument with alias {inst_alias} could not be found within the instruments of the "
                        "platform. The available instrument aliases are: "
                        f"{[inst.alias for inst in platform_instruments.elements]}."
                    )
                instruments.append(inst_class)
            self.instruments = instruments
            super().__post_init__()

    settings: SystemControlSettings

    def __init__(self, settings: dict, platform_instruments: Instruments | None = None):
        settings_class: type[self.SystemControlSettings] = get_type_hints(self).get("settings")  # type: ignore
        self.settings = settings_class(**settings, platform_instruments=platform_instruments)

    def compile(
        self, pulse_bus_schedule: PulseBusSchedule, nshots: int, repetition_duration: int, num_bins: int
    ) -> list:
        """Compiles the ``PulseBusSchedule`` into an assembly program.

        Args:
            pulse_bus_schedule (PulseBusSchedule): the list of pulses to be converted into a program
            nshots (int): number of shots / hardware average
            repetition_duration (int): maximum window for the duration of one hardware repetition
            num_bins (int): number of bins
        """
        for instrument in self.instruments:
            if isinstance(instrument, AWG):
                return instrument.compile(
                    pulse_bus_schedule=pulse_bus_schedule,
                    nshots=nshots,
                    repetition_duration=repetition_duration,
                    num_bins=num_bins,
                )
        raise AttributeError("The system control doesn't have any AWG to compile the given pulse sequence.")

    def upload_qpysequence(self, qpysequence: QpySequence, port: str):
        """Uploads the qpysequence into the instrument."""
        for instrument in self.instruments:
            if isinstance(instrument, AWG):
                instrument.upload_qpysequence(qpysequence=qpysequence, port=port)
                return

        raise AttributeError("The system control doesn't have any AWG to upload a qpysequence.")

    def upload(self, port: str):
        """Uploads any previously compiled program into the instrument."""
        for instrument in self.instruments:
            if isinstance(instrument, AWG):
                instrument.upload(port=port)
                return

        raise AttributeError("The system control doesn't have any AWG to upload a program.")

    def run(self, port: str) -> None:
        """Runs any previously uploaded program into the instrument."""
        for instrument in self.instruments:
            if isinstance(instrument, AWG):
                instrument.run(port=port)
                return

        raise AttributeError("The system control doesn't have any AWG to run a program.")

    def __str__(self):
        """String representation of a SystemControl class."""
        return "".join(f"-|{instrument}|-" for instrument in self.instruments)

    def __iter__(self):
        """Redirect __iter__ magic method."""
        return iter(self.settings.instruments)

    def to_dict(self):
        """Return a dict representation of a SystemControl class."""
        return {RUNCARD.NAME: self.name.value, RUNCARD.INSTRUMENTS: [inst.alias for inst in self.instruments]}

    @property
    def instruments(self) -> list[Instrument]:
        """Instruments controlled by this system control."""
        return self.settings.instruments

    def set_parameter(
        self, parameter: Parameter, value: float | str | bool, channel_id: int | None = None, port_id: str | None = None
    ):
        """Sets the parameter of a specific instrument.

        Args:
            parameter (Parameter): parameter settings of the instrument to update
            value (float | str | bool): value to update
            channel_id (int | None, optional): instrument channel to update, if multiple. Defaults to None.
            port_id (str | None, optional): The ``port_id`` argument can be used when setting a parameter of a
                QbloxModule, to avoid having to look which sequencer corresponds to which bus.
        """
        for instrument in self.instruments:
            with contextlib.suppress(ParameterNotFound):
                if isinstance(instrument, QbloxModule):
                    instrument.setup(parameter, value, channel_id, port_id=port_id)
                else:
                    instrument.set_parameter(parameter, value, channel_id)
                return
        raise ParameterNotFound(f"Could not find parameter {parameter.value} in the system control {self.name}")

    def get_parameter(self, parameter: Parameter, channel_id: int | None = None, port_id: str | None = None):
        """Gets a parameter of a specific instrument.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None, optional): instrument channel to update, if multiple. Defaults to None.
        """
        for instrument in self.instruments:
            with contextlib.suppress(ParameterNotFound):
                if isinstance(instrument, QbloxModule):
                    return instrument.get(parameter, channel_id, port_id=port_id)
                return instrument.get_parameter(parameter, channel_id)
        raise ParameterNotFound(f"Could not find parameter {parameter.value} in the system control {self.name}")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Class to interface with the voltage source Qblox S4g
"""
from dataclasses import dataclass
from time import sleep
from typing import Any

from qililab.config import logger
from qililab.instruments.current_source import CurrentSource
from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.instruments.utils import InstrumentFactory
from qililab.typings import InstrumentName
from qililab.typings import QbloxS4g as QbloxS4gDriver
from qililab.typings.enums import Parameter


@InstrumentFactory.register
class QbloxS4g(CurrentSource):
    """Qblox S4g class

    Args:
        name (InstrumentName): name of the instrument
        device (Qblox_S4g): Instance of the qcodes S4g class.
        settings (QbloxS4gSettings): Settings of the instrument.
    """

    name = InstrumentName.QBLOX_S4G

    @dataclass
    class QbloxS4gSettings(CurrentSource.CurrentSourceSettings):
        """Contains the settings of a specific signal generator."""

    settings: QbloxS4gSettings
    device: QbloxS4gDriver

    def dac(self, dac_index: int):
        """get channel associated to the specific dac

        Args:
            dac_index (int): channel index

        Returns:
            _type_: _description_
        """
        return getattr(self.device, f"dac{dac_index}")

    def _channel_setup(self, dac_index: int) -> None:
        """Setup for a specific dac channel

        Args:
            dac_index (int): dac specific index channel
        """
        channel = self.dac(dac_index=dac_index)
        channel.ramping_enabled(self.ramping_enabled[dac_index])
        channel.ramp_rate(self.ramp_rate[dac_index])
        channel.span(self.span[dac_index])
        channel.current(self.current[dac_index])
        logger.debug("SPI current set to %f", channel.current())
        while channel.is_ramping():
            sleep(0.1)

    @Instrument.CheckDeviceInitialized
    def setup(self, parameter: Parameter, value: float | str | bool, channel_id: int | None = None):
        """Set Qblox instrument calibration settings."""

        if channel_id is None:
            if len(self.dacs) == 1:
                channel_id = self.dacs[0]
            else:
                raise ValueError(f"channel not specified to update instrument {self.name.value}")
        if channel_id > 3:
            raise ValueError(
                f"the specified dac index:{channel_id} is out of range."
                + " Number of dacs is 4 -> maximum channel_id should be 3."
            )
        channel = self.dac(dac_index=channel_id)
        if parameter == Parameter.CURRENT:
            self._set_current(value=value, channel_id=channel_id, channel=channel)
            return
        if parameter == Parameter.SPAN:
            self._set_span(value=value, channel_id=channel_id, channel=channel)
            return
        if parameter == Parameter.RAMPING_ENABLED:
            self._set_ramping_enabled(value=value, channel_id=channel_id, channel=channel)
            return
        if parameter == Parameter.RAMPING_RATE:
            self._set_ramping_rate(value=value, channel_id=channel_id, channel=channel)
            return
        raise ParameterNotFound(f"Invalid Parameter: {parameter.value}")

    def get(self, parameter: Parameter, channel_id: int | None = None):
        """Get instrument parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None): Channel identifier of the parameter to update.
        """
        if channel_id is None:
            if len(self.dacs) == 1:
                channel_id = self.dacs[0]
            else:
                raise ValueError(f"channel not specified to update instrument {self.name.value}")
        if channel_id > 3:
            raise ValueError(
                f"the specified dac index:{channel_id} is out of range."
                + " Number of dacs is 4 -> maximum channel_id should be 3."
            )
        if hasattr(self.settings, parameter.value):
            return getattr(self.settings, parameter.value)[channel_id]
        raise ParameterNotFound(f"Could not find parameter {parameter} in instrument {self.name}")

    @Instrument.CheckParameterValueFloatOrInt
    def _set_current(self, value: float | str | bool, channel_id: int, channel: Any):
        """Set the current"""
        self.settings.current[channel_id] = float(value)
        channel.current(self.current[channel_id])

    @Instrument.CheckParameterValueString
    def _set_span(self, value: float | str | bool, channel_id: int, channel: Any):
        """Set the span"""
        self.settings.span[channel_id] = str(value)
        channel.span(self.span[channel_id])

    @Instrument.CheckParameterValueBool
    def _set_ramping_enabled(self, value: float | str | bool, channel_id: int, channel: Any):
        """Set the ramping_enabled"""
        self.settings.ramping_enabled[channel_id] = bool(value)
        channel.ramping_enabled(self.ramping_enabled[channel_id])

    @Instrument.CheckParameterValueFloatOrInt
    def _set_ramping_rate(self, value: float | str | bool, channel_id: int, channel: Any):
        """Set the ramp_rate"""
        self.settings.ramp_rate[channel_id] = float(value)
        channel.ramp_rate(self.ramp_rate[channel_id])

    @Instrument.CheckDeviceInitialized
    def initial_setup(self):
        """performs an initial setup."""
        for dac_index in self.dacs:
            self._channel_setup(dac_index=dac_index)

    @Instrument.CheckDeviceInitialized
    def turn_on(self):
        """Dummy method."""

    @Instrument.CheckDeviceInitialized
    def turn_off(self):
        """Stop outputing current."""
        self.device.set_dacs_zero()
        for dac_index in self.settings.dacs:
            channel = self.dac(dac_index=dac_index)
            logger.debug("Dac%d current resetted to  %f", dac_index, channel.current())

    @Instrument.CheckDeviceInitialized
    def reset(self):
        """Reset instrument."""
        self.device.set_dacs_zero()



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""__init__.py"""
from .awg import AWG
from .awg_analog_digital_converter import AWGAnalogDigitalConverter
from .instrument import Instrument, ParameterNotFound
from .instruments import Instruments
from .mini_circuits import Attenuator
from .quantum_machines import QuantumMachinesManager
from .rohde_schwarz import SGS100A
from .signal_generator import SignalGenerator
from .utils import InstrumentFactory



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Arbitrary waveform."""
import numpy as np

from .waveform import Waveform


class Arbitrary(Waveform):  # pylint: disable=too-few-public-methods, disable=missing-class-docstring
    """Arbitrary waveform. Creates a waveform with the passed envelope.

    Args:
        samples(np.ndarray): Passed envelope to base the waveform on.

    Examples:
        If you want to create a waveform with an envelope given by:

        .. code-block:: python

            import numpy as np
            samples = np.ones(50)

        You would just need to do:

        .. code-block:: python

            import qililab as ql
            arbitrary_envelope = ql.Arbitrary(samples=samples)
    """

    def __init__(self, samples: np.ndarray):
        self.samples = samples

    def envelope(self, resolution: int = 1) -> np.ndarray:
        """Returns the envelope corresponding to the arbitrary waveform.

        Args:
            resolution (int, optional): Pulse resolution. Defaults to 1.

        Returns:
            np.ndarray: Height of the envelope for each time step.
        """
        if resolution == 1:
            return self.samples

        # Calculate the averaging window
        averaging_window = int(len(self.samples) / resolution)

        # Calculate the moving average
        cumsum = np.cumsum(self.samples)
        cumsum[averaging_window:] = cumsum[averaging_window:] - cumsum[:-averaging_window]
        moving_avg = cumsum[averaging_window:] / averaging_window

        # Apply scaling and shifting
        scale_factor = (self.samples.max() - self.samples.min()) / (moving_avg.max() - moving_avg.min())
        shift = self.samples.max() - moving_avg.max() * scale_factor
        scaled_avg = moving_avg * scale_factor + shift

        # Apply clamping
        clamped_avg = np.clip(scaled_avg, self.samples.min(), self.samples.max())

        return clamped_avg
